
import pytest
from rocketry.core.time import (
    TimeDelta
)
from rocketry.pybox.time.convert import to_datetime

@pytest.mark.parametrize(
    "dt,past,future,roll_start,roll_end",
    [
        # Regular
        pytest.param(
            to_datetime("2020-01-01 10:00:00"),
            None, None,
            to_datetime("2020-01-01 10:00:00"), to_datetime("2020-01-01 10:00:00"),
            id="No roll"),
        pytest.param(
            to_datetime("2020-01-01 10:00:00"),
            None, "1:20:30",
            to_datetime("2020-01-01 10:00:00"), to_datetime("2020-01-01 11:20:30"),
            id="Regular"),
    ],
)
def test_rollforward(dt, past, future, roll_start, roll_end):
    time = TimeDelta(past, future)

    interval = time.rollforward(dt)
    assert roll_start == interval.left
    assert roll_end == interval.right


@pytest.mark.parametrize(
    "dt,past,future,roll_start,roll_end",
    [
        # Regular
        pytest.param(
            to_datetime("2020-01-01 10:00:00"),
            None, None,
            to_datetime("2020-01-01 10:00:00"), to_datetime("2020-01-01 10:00:00"),
            id="No roll"),
        pytest.param(
            to_datetime("2020-01-01 10:00:00"),
            "1:20:30", None,
            to_datetime("2020-01-01 08:39:30"), to_datetime("2020-01-01 10:00:00"),
            id="Regular"),
    ],
)
def test_rollback(dt, past, future, roll_start, roll_end):
    time = TimeDelta(past, future)

    interval = time.rollback(dt)
    assert roll_start == interval.left
    assert roll_end == interval.right
    



import subprocess
from typing import List, Optional, Union

try:
    from typing import Literal
except ImportError: # pragma: no cover
    from typing_extensions import Literal

from pydantic import Field, validator

from rocketry.core.parameters.parameters import Parameters
from rocketry.core.task import Task


class CommandTask(Task):
    """Task that executes a command from
    shell/terminal.

    Parameters
    ----------
    command : str, list
        Command to execute.
    cwd : str, optional
        Sets the current directory before the child is executed.
    shell : bool, optional
        If true, the command will be executed through the shell.
    kwds_popen : dict, optional
        Keyword arguments to be passed to subprocess.Popen
    **kwargs : dict
        See :py:class:`rocketry.core.Task`

    Examples
    --------

    >>> from rocketry.tasks import CommandTask
    >>> task = CommandTask("python -m pip install rocketry", name="my_cmd_task_1")

    Or list of commands:

    >>> task = CommandTask(["python", "-m", "pip", "install", "rocketry"], name="my_cmd_task_2")
    """

    command: Union[str, List[str]]
    shell: bool = False
    cwd: Optional[str]
    kwds_popen: dict = {}
    argform: Optional[Literal['-', '--', 'short', 'long']] = Field(description="Whether the arguments are turned as short or long form command line arguments")

    def get_kwargs_popen(self) -> dict:
        kwargs = {
            "cwd": self.cwd,
            "shell": self.shell,
            "stdin": subprocess.PIPE,
            "stdout": subprocess.PIPE,
            "stderr": subprocess.PIPE,
        }
        kwargs.update(self.kwds_popen)
        return kwargs

    @validator('argform')
    def parse_argform(cls, value):
        return {
            "long": "--",
            "--": "--",
            "short": "-",
            "-": "-",
            None: '--',
        }[value]

    def execute(self, **parameters):
        """Run the command."""
        command = self.command

        for param, val in parameters.items():
            if not param.startswith("-"):
                param = self.argform + param

            if isinstance(command, str):
                command = command + f" {param} \"{val}\""
            else:
                command += [param] + [val]

        # https://stackoverflow.com/a/5469427/13696660
        pipe = subprocess.Popen(command, **self.get_kwargs_popen())
        try:
            outs, errs = pipe.communicate(timeout=self.timeout)
        except subprocess.TimeoutExpired:
            # https://docs.python.org/3.3/library/subprocess.html#subprocess.Popen.communicate
            pipe.kill()
            outs, errs = pipe.communicate()
            raise

        return_code = pipe.returncode
        if return_code != 0:
            if hasattr(errs, "decode"):
                errs = errs.decode("utf-8", errors="ignore")
            raise OSError(f"Failed running command ({return_code}): \n{errs}")
        return outs

    def postfilter_params(self, params: Parameters):
        # Only allows the task specific parameters
        # for simplicity
        return params

    def get_default_name(self, command, **kwargs):
        return command if isinstance(command, str) else ' '.join(command)



from copy import copy
from abc import abstractmethod
from typing import Callable, Dict, Pattern, Union

from rocketry._base import RedBase
from rocketry.core.parameters.parameters import Parameters

PARSERS: Dict[Union[str, Pattern], Union[Callable, 'BaseCondition']] = {}


class BaseCondition(RedBase):
    """A condition is a thing/occurence that should happen in
    order to something happen.

    Conditions are used to determine whether a task can be started,
    a task should be terminated or the scheduler should shut
    down. Conditions are either true or false.

    A condition could answer for any of the following questions:
        - Current time is as specified (ie. Monday afternoon).
        - A given task has already run.
        - The machine has at least a given amount of RAM.
        - A specific file exists.

    Each condition should have the method ``__bool__`` specified
    as minimum. This method should return ``True`` or ``False``
    depending on whether the condition holds or does not hold.

    Examples
    --------

    Minimum example:

    >>> from rocketry.core import BaseCondition
    >>> class MyCondition(BaseCondition):
    ...     def __bool__(self):
    ...         ... # Code that defines state either
    ...         return True

    Complicated example with parser:

    >>> import os, re
    >>> class IsFooBar(BaseCondition):
    ...     __parsers__ = {
    ...         re.compile(r"is foo '(?P<outcome>.+)'"): "__init__"
    ...     }
    ...
    ...     def __init__(self, outcome):
    ...         self.outcome = outcome
    ...
    ...     def __bool__(self):
    ...         return self.outcome == "bar"
    ...
    ...     def __repr__(self):
    ...         return f"IsFooBar('{self.outcome}')"
    ...
    >>> from rocketry.parse import parse_condition
    >>> parse_condition("is foo 'bar'")
    IsFooBar('bar')

    """

    def observe(self, **kwargs):
        "Observe the status of the condition"
        cond_params = Parameters._from_signature(self.get_state, **kwargs)
        param_dict = cond_params.materialize(**kwargs)
        return self.get_state(**param_dict)

    def __bool__(self) -> bool:
        """Check whether the condition holds."""
        return self.observe()

    @abstractmethod
    def get_state(self):
        """Get the status of the condition
        (using arguments)

        Override this method."""

    def __and__(self, other):
        # self & other
        # bitwise and
        # using & operator

        return All(self, other)

    def __or__(self, other):
        # self | other
        # bitwise or

        return Any(self, other)

    def __invert__(self):
        # ~self
        # bitwise not
        return Not(self)

    def __eq__(self, other):
        "Equal operation"
        is_same_class = isinstance(other, type(self))
        if is_same_class:
            # Check equality of the attributes except
            # those that are only for display purposes
            repr_attrs = ("_str",)
            self_dict = {
                key: val for key, val in self.__dict__.items()
                if key not in repr_attrs
            }
            other_dict = {
                key: val for key, val in other.__dict__.items()
                if key not in repr_attrs
            }
            return self_dict == other_dict
        return False

    def __str__(self):
        if hasattr(self, "_str"):
            return self._str
        else:
            raise AttributeError(f"Condition {type(self)} is missing __str__.")


class _ConditionContainer:
    "Wraps another condition"

    def __getitem__(self, val):
        return self.subconditions[val]

    def __iter__(self):
        return iter(self.subconditions)

    def __eq__(self, other):
        "Equal operation"
        is_same_class = isinstance(other, type(self))
        if is_same_class:
            return self.subconditions == other.subconditions
        else:
            return False

    def __repr__(self):
        string = ', '.join(map(str, self.subconditions))
        return f'{type(self).__name__}({string})'

class Any(_ConditionContainer, BaseCondition):

    def __init__(self, *conditions):
        self.subconditions = []

        self_type = type(self)
        for cond in conditions:
            # Avoiding nesting (like Any(Any(...), ...) --> Any(...))
            conds = cond.subconditions if isinstance(cond, self_type) else [cond]
            self.subconditions += conds

    def observe(self, **kwargs) -> bool:
        for subcond in self.subconditions:
            if subcond.observe(**kwargs):
                return True
        return False

    def __str__(self):
        try:
            return super().__str__()
        except AttributeError:
            string = ' | '.join(map(str, self.subconditions))
            return f'({string})'


class All(_ConditionContainer, BaseCondition):

    def __init__(self, *conditions):
        self.subconditions = []

        self_type = type(self)
        for cond in conditions:
            # Avoiding nesting (like All(All(...), ...) --> All(...))
            conds = cond.subconditions if isinstance(cond, self_type) else [cond]
            self.subconditions += conds

    def observe(self, **kwargs) -> bool:
        for subcond in self.subconditions:
            if not subcond.observe(**kwargs):
                return False
        return True

    def __str__(self):
        try:
            return super().__str__()
        except AttributeError:
            string = ' & '.join(map(str, self.subconditions))
            return f'({string})'


class Not(_ConditionContainer, BaseCondition):

    def __init__(self, condition):
        # TODO: rename condition as child
        self.condition = condition

    def observe(self, **kwargs):
        return not self.condition.observe(**kwargs)

    def __repr__(self):
        string = repr(self.condition)
        return f'Not({string})'

    def __str__(self):
        try:
            return super().__str__()
        except AttributeError:
            string = str(self.condition)
            return f'~{string}'

    @property
    def subconditions(self):
        return (self.condition,)

    def __invert__(self):
        "inverse of inverse is the actual condition"
        return self.condition

    def __eq__(self, other):
        "Equal operation"
        if isinstance(other, AlwaysTrue):
            return isinstance(self.condition, AlwaysFalse)
        elif isinstance(other, AlwaysFalse):
            return isinstance(self.condition, AlwaysTrue)

        is_same_class = isinstance(other, type(self))
        if is_same_class:
            return self.condition == other.condition
        else:
            return False


class AlwaysTrue(BaseCondition):
    "Condition that is always true"
    def observe(self, **kwargs):
        return True

    def __repr__(self):
        return 'true'

    def __str__(self):
        return 'true'

class AlwaysFalse(BaseCondition):
    "Condition that is always false"

    def observe(self, **kwargs):
        return False

    def __repr__(self):
        return 'false'

    def __str__(self):
        return 'false'


class BaseComparable(BaseCondition):

    _comp_attrs = ("__eq__", "__ne__", "__lt__", "__gt__", "__le__", "__ge__")

    def __init__(self):
        self._comps = {}
        super().__init__()

    def observe(self, **kwargs):
        params = Parameters._from_signature(self.get_measurement, **kwargs)
        param_dict = params.materialize(**kwargs)
        value = self.get_measurement(**param_dict)
        if isinstance(value, bool):
            # Possibly has some optimization and already did the comparison
            return value
        return self.get_state(value)

    @abstractmethod
    def get_measurement(self):
        "Get measurement (something that can be compared)"

    def get_state(self, res:int):
        compares = self._comps

        res = len(res) if hasattr(res, "__len__") else res

        if not compares:
            return res > 0
        return all(
            getattr(res, comp)(val) # Comparison is magic method (==, !=, etc.)
            for comp, val in compares.items()
        )

    def _is_any_over_zero(self):
        # Useful for optimization: just find any observation and the statement is true
        comps = {
            comp: self._comps[comp]
            for comp in self._comp_attrs
            if comp in self._comps
        }
        if comps == {"__gt__": 0} or comps == {"__ge__": 1} or comps == {"__gt__": 0, "__ge__": 1}:
            return True
        return not comps

    def _is_equal_zero(self):
        comps = {
            comp: self._comps[comp]
            for comp in self._comp_attrs
            if comp in self._comps
        }
        return comps == {"__eq__": 0}

    def __eq__(self, other):
        # self == other
        is_same_class = isinstance(other, BaseCondition)
        if is_same_class:
            # Not storing as parameter to statement but
            # check whether the statements are same
            return super().__eq__(other)
        return self._set_comparison("__eq__", other)

    def __ne__(self, other):
        # self != other
        return self._set_comparison("__ne__", other)

    def __lt__(self, other):
        # self < other
        return self._set_comparison("__lt__", other)

    def __gt__(self, other):
        # self > other
        return self._set_comparison("__gt__", other)

    def __le__(self, other):
        # self <= other
        return self._set_comparison("__le__", other)

    def __ge__(self, other):
        # self >= other
        return self._set_comparison("__ge__", other)

    def _set_comparison(self, key, val):
        obj = copy(self)
        obj._comps[key] = val
        return obj

    @classmethod
    def from_magic(cls, **kwargs):
        for key in kwargs:
            if key not in cls._comp_attrs:
                raise ValueError(f"Unknown comparison: {key}")
        obj = cls()
        obj._comps = kwargs
        return obj



import datetime

import pytest

from rocketry.core.time.base import StaticInterval
from rocketry.parse import parse_time, ParserError
from rocketry.time import TimeOfDay, TimeOfHour, TimeOfMinute, TimeDelta, TimeOfWeek, TimeOfMonth, always, never

@pytest.mark.parametrize(
    "time_str,expected", [
        pytest.param("always", always, id="always"),
        pytest.param("never", never, id="never"),

        pytest.param("every 10 seconds", TimeDelta("10 seconds"), id="every"),

        pytest.param("time of minute between 15:00 and 30:00", TimeOfMinute("15:00", "30:00"), id="TimeOfMinute between"),
        pytest.param("time of minute before 30:00", TimeOfMinute(None, "30:00"), id="TimeOfMinute before"),
        pytest.param("time of minute after 30:00", TimeOfMinute("30:00", None), id="TimeOfMinute after"),

        pytest.param("time of hour between 15:00 and 30:00", TimeOfHour("15:00", "30:00"), id="TimeOfHour between"),
        pytest.param("time of hour before 30:00", TimeOfHour(None, "30:00"), id="TimeOfHour before"),
        pytest.param("time of hour after 30:00", TimeOfHour("30:00", None), id="TimeOfHour after"),

        pytest.param("time of day between 10:00 and 12:00", TimeOfDay("10:00", "12:00"), id="TimeOfDay between"),
        pytest.param("time of day before 12:00", TimeOfDay(None, "12:00"), id="TimeOfDay before"),
        pytest.param("time of day after 12:00", TimeOfDay("12:00", None), id="TimeOfDay after"),

        pytest.param("time of week on Monday", TimeOfWeek.at("Mon"), id="TimeOfWeek on"),

        pytest.param("time of month between 1st and 2nd", TimeOfMonth("1st", "2nd"), id="TimeOfMonth between (1st, 2nd)"),
        pytest.param("time of month between 3rd and 4th", TimeOfMonth("3rd", "4th"), id="TimeOfMonth between (3rd, 4th)"),

        pytest.param("time of week on Monday | time of week on Friday", TimeOfWeek.at("Mon") | TimeOfWeek.at("Fri"), id="OR"),
        pytest.param("time of week on Monday & time of day between 10:00 and 12:00", TimeOfDay("10:00", "12:00") & TimeOfWeek.at("Mon"), id="AND"),
    ]
)
def test_string(time_str, expected):
    cond = parse_time(time_str)
    assert cond == expected

@pytest.mark.parametrize(
    "t,s,r", [
        pytest.param(always, "always", "always", id="always"),
        pytest.param(never, "never", "never", id="never"),
        pytest.param(
            StaticInterval(datetime.datetime(2022, 1, 1), datetime.datetime(2022, 12, 31)),
            "between 2022-01-01 00:00:00 and 2022-12-31 00:00:00",
            "StaticInterval(start=datetime.datetime(2022, 1, 1, 0, 0), end=datetime.datetime(2022, 12, 31, 0, 0))",
            id="static interval"
        ),
    ]
)
def test_representation(t, s, r):
    assert str(t) == s
    assert repr(t) == r

def test_error():
    with pytest.raises(ParserError):
        cond = parse_time("invalid string")



# Imports
from .base import (
    TimePeriod, TimeInterval,
    TimeDelta,
    StaticInterval,
    All, Any,

    PARSERS,

    # Constants
    always,
    never,
)


#from .factory import period_factory



import datetime

from rocketry.core.time.base import (
    All, Any
)
from rocketry.time import TimeOfDay, TimeOfMinute, always

from_iso = datetime.datetime.fromisoformat

def test_compress_any():
    assert Any(
        TimeOfDay(),
        TimeOfDay(),
        Any(TimeOfMinute(), TimeOfMinute()), # This is unnested
        All(TimeOfDay(), TimeOfMinute()),
    ) == Any(
        TimeOfDay(),
        TimeOfDay(),
        TimeOfMinute(), TimeOfMinute(), # This was unnested
        All(TimeOfDay(), TimeOfMinute()),
    )

    assert Any(
        TimeOfDay(),
        All(TimeOfDay(), TimeOfMinute()),
        always, # The other periods does not matter
        Any(TimeOfMinute(), TimeOfMinute()),
    ) == Any(always)


def test_compress_all():
    assert All(
        TimeOfDay(),
        TimeOfDay(),
        All(TimeOfMinute(), TimeOfMinute()), # This is unnested
        Any(TimeOfDay(), TimeOfMinute()),
    ) == All(
        TimeOfDay(),
        TimeOfDay(),
        TimeOfMinute(), TimeOfMinute(), # This was unnested
        Any(TimeOfDay(), TimeOfMinute()),
    )

    assert All(
        TimeOfDay(),
        All(TimeOfDay(), TimeOfMinute()),
        always, # This has no effect
        Any(TimeOfMinute(), TimeOfMinute()),
    ) == All(
        TimeOfDay(),
        All(TimeOfDay(), TimeOfMinute()),
        Any(TimeOfMinute(), TimeOfMinute()),
    )



import asyncio
import logging
import pytest

from redbird.logging import RepoHandler
from redbird.repos import MemoryRepo, CSVFileRepo

from rocketry import Rocketry
from rocketry.conditions.task.task import TaskStarted
from rocketry.args import Return, Arg, FuncArg
from rocketry import Session
from rocketry.tasks import CommandTask
from rocketry.tasks import FuncTask
from rocketry.conds import false, true

def set_logging_defaults():
    task_logger = logging.getLogger("rocketry.task")
    task_logger.handlers = []
    task_logger.setLevel(logging.WARNING)

def test_app_create(session, tmpdir):
    set_logging_defaults()

    with pytest.warns(FutureWarning):
        app = Rocketry()

    # Test logging
    task_logger = logging.getLogger("rocketry.task")

    # Till Red Bird supports equal, we need to test the handler one obj at a time
    assert len(task_logger.handlers) == 1
    assert isinstance(task_logger.handlers[0], RepoHandler)
    assert isinstance(task_logger.handlers[0].repo, MemoryRepo)

    assert isinstance(app.session, Session)

    # Test setting SQL repo
    with pytest.warns(FutureWarning):
        with tmpdir.as_cwd():
            app = Rocketry(logger_repo=CSVFileRepo(filename="myrepo.csv"))
    assert len(task_logger.handlers) == 2
    assert isinstance(task_logger.handlers[0], RepoHandler)
    assert isinstance(task_logger.handlers[0].repo, CSVFileRepo)

    assert isinstance(app.session, Session)

    app = Rocketry(execution="thread")
    assert app.session.config.task_execution == "thread"

def test_app():
    set_logging_defaults()

    app = Rocketry(execution="thread")
    assert app.session.config.task_execution == "thread"

def test_app_tasks():
    set_logging_defaults()

    app = Rocketry(config={'task_execution': 'main'})

    # Creating some tasks
    @app.task()
    def do_never():
        ...

    @app.task('daily')
    def do_func():
        ...
        return 'return value'

    app.task('daily', name="do_command", command="echo 'hello world'")
    app.task('daily', name="do_script", path=__file__)

    # Assert and test tasks
    assert len(app.session.tasks) == 4

    assert isinstance(app.session['do_func'], FuncTask)
    assert isinstance(app.session['do_command'], CommandTask)
    assert isinstance(app.session['do_script'], FuncTask)

    assert app.session['do_never'].start_cond == false

def test_nested_args():
    set_logging_defaults()

    # Creating app
    app = Rocketry(config={'task_execution': 'main'})

    @app.param('arg_1')
    def my_arg_1():
        return 'arg 1'

    @app.param('arg_2')
    def my_func_2(arg=Arg('arg_1')):
        assert arg == "arg 1"
        return 'arg 2'

    @app.param('arg_3')
    def my_func_3(arg_1=Arg('arg_1'), arg_2=Arg("arg_2")):
        assert arg_1 == "arg 1"
        assert arg_2 == "arg 2"
        return 'arg 3'

    # Creating a task to test this
    @app.task(true)
    def do_daily(arg=Arg('arg_3')):
        ...
        assert arg == "arg 3"

    app.session.config.shut_cond = TaskStarted(task='do_daily')
    app.run()
    logger = app.session['do_daily'].logger
    assert logger.filter_by(action="success").count() == 1

def test_nested_args_from_func_arg():
    set_logging_defaults()

    # Creating app
    app = Rocketry(config={'task_execution': 'main'})

    @app.param('arg_1')
    def my_arg_1():
        return 'arg 1'

    def my_func_2(arg=Arg('arg_1')):
        assert arg == "arg 1"
        return 'arg 2'

    def my_func_3(arg_1=Arg('arg_1'), arg_2=FuncArg(my_func_2)):
        assert arg_1 == "arg 1"
        assert arg_2 == "arg 2"
        return 'arg 3'

    # Creating a task to test this
    @app.task(true)
    def do_daily(arg=FuncArg(my_func_3)):
        ...
        assert arg == "arg 3"

    app.session.config.shut_cond = TaskStarted(task='do_daily')
    app.run()
    logger = app.session['do_daily'].logger
    assert logger.filter_by(action="success").count() == 1

def test_arg_ref():
    set_logging_defaults()

    # Creating app
    app = Rocketry(config={'task_execution': 'main'})

    @app.param('arg_1')
    def my_arg_1():
        return 'arg 1'

    @app.param('arg_2')
    def my_arg_2():
        return 'arg 2'

    # Creating a task to test this
    @app.task(true)
    def do_daily(arg_1=Arg(my_arg_1), arg_2=Arg(my_arg_2)):
        ...
        assert arg_1 == "arg 1"
        assert arg_2 == "arg 2"

    app.session.config.shut_cond = TaskStarted(task='do_daily')
    app.run()
    logger = app.session['do_daily'].logger
    assert logger.filter_by(action="success").count() == 1

def test_app_async():
    set_logging_defaults()

    app = Rocketry(config={'task_execution': 'main'})

    # Creating some tasks
    @app.task()
    def do_never():
        ...

    @app.task('daily')
    def do_func():
        ...
        return 'return value'

    app.session.config.shut_cond = TaskStarted(task='do_func')
    asyncio.run(app.serve())
    assert 1 == app.session[do_func].logger.filter_by(action="success").count()

def test_app_run():
    set_logging_defaults()

    # Creating app
    app = Rocketry(config={'task_execution': 'main'})
    app.params(my_arg='session value')

    # Creating some params and conditions
    @app.param('my_func_arg')
    def my_func():
        return 'func arg value'

    @app.cond('is foo')
    def is_foo(arg=Arg("my_func_arg")):
        assert arg == "func arg value"
        return True

    # Creating some tasks
    @app.task('daily & is foo')
    def do_daily():
        ...
        return 'return value'

    @app.task("after task 'do_daily'")
    def do_after(arg=Return('do_daily'), session_arg=Arg('my_arg'), func_arg=Arg('my_func_arg'), func_arg_2=FuncArg(lambda: 'my val')):
        assert arg == 'return value'
        assert session_arg == 'session value'
        assert func_arg == 'func arg value'
        assert func_arg_2 == 'my val'

    @app.task("true")
    def do_fail():
        raise

    @app.task("false", execution='process', name="never done", parameters={'arg_1': 'something'})
    def do_never(arg_1):
        raise

    app.session.config.shut_cond = TaskStarted(task='do_after')

    # Testing the app has all relevant things
    session = app.session
    assert 'is foo' in list(session._cond_parsers)
    assert set(task.name for task in session.tasks) == {'do_daily', 'do_after', 'do_fail', 'never done'}

    app.run()

    logger = app.session['do_after'].logger
    assert logger.filter_by(action="success").count() == 1

    logger = app.session['do_fail'].logger
    assert logger.filter_by(action="fail").count() == 1

    # Test the parameters are passed to a task
    task_example = session['never done']
    assert task_example.execution == 'process'
    assert task_example.name == 'never done'
    assert dict(task_example.parameters) == {'arg_1': 'something'}


def test_task_name():
    set_logging_defaults()

    app = Rocketry(config={'task_execution': 'main'})

    @app.task()
    def do_func():
        ...
        return 'return value'

    assert app.session[do_func].name == "do_func"



from datetime import datetime
import pytest
from rocketry.pybox.time import Interval, to_datetime

@pytest.mark.parametrize("l,r",
    [
        pytest.param(
            Interval(to_datetime("2022-07-01"), to_datetime("2022-07-30")),
            Interval(to_datetime("2022-07-12"), to_datetime("2022-07-15")),
            id="right inside left"
        ),
        pytest.param(
            Interval(to_datetime("2022-07-12"), to_datetime("2022-07-15")),
            Interval(to_datetime("2022-07-01"), to_datetime("2022-07-30")),
            id="left inside right"
        ),

        pytest.param(
            Interval(to_datetime("2022-07-01"), to_datetime("2022-07-15"), closed="neither"),
            Interval(to_datetime("2022-07-14"), to_datetime("2022-07-30"), closed="neither"),
            id="left partially right (right)"
        ),
        pytest.param(
            Interval(to_datetime("2022-07-14"), to_datetime("2022-07-30"), closed="neither"),
            Interval(to_datetime("2022-07-01"), to_datetime("2022-07-15"), closed="neither"),
            id="left partially right (left)"
        ),

        pytest.param(
            Interval(to_datetime("2022-07-01"), to_datetime("2022-07-12"), closed="both"),
            Interval(to_datetime("2022-07-12"), to_datetime("2022-07-15"), closed="both"),
            id=f"left edges right (right)"
        ),
        pytest.param(
            Interval(to_datetime("2022-07-12"), to_datetime("2022-07-15"), closed="both"),
            Interval(to_datetime("2022-07-01"), to_datetime("2022-07-12"), closed="both"),
            id=f"left edges right (left)"
        ),
    ]
)
def test_overlaps(l, r):
    assert l.overlaps(r)


@pytest.mark.parametrize("l,r",
    [
        pytest.param(
            Interval(to_datetime("2022-07-01"), to_datetime("2022-07-12"), closed=l_closed),
            Interval(to_datetime("2022-07-12"), to_datetime("2022-07-30"), closed=r_closed),
            id=f"right edge ({l_closed}, {r_closed})"
        )
        for l_closed, r_closed in [
            ("left", 'left'), ('left', 'both'), ('left', 'neither'),
            ("left", 'right'), ('both', 'right'), ('neither', 'right'),
            ("neither", 'both'), ('neither', 'neither'),
        ]
    ]
)
def test_not_overlaps(l, r):
    assert not l.overlaps(r)

def test_fail():
    with pytest.raises(ValueError):
        Interval(datetime(2022, 1, 1), datetime(2019, 1, 1))
    with pytest.raises(ValueError):
        Interval(datetime(2022, 1, 1), datetime(2022, 1, 2), closed="typo")

def test_empty():
    assert not Interval(datetime(2022, 1, 1), datetime(2022, 1, 1), closed="both").is_empty
    for closed in ("left", "right", "both", "neither"):
        assert not Interval(datetime(2022, 1, 1), datetime(2022, 1, 2), closed=closed).is_empty
    for closed in ("left", "right", "neither"):
        assert Interval(datetime(2022, 1, 1), datetime(2022, 1, 1), closed=closed).is_empty

def test_repr():
    for closed in ("left", "right", "neither"):
        assert repr(Interval(datetime(2022, 1, 1), datetime(2022, 1, 1), closed=closed))



from typing import Callable
from rocketry.args import FuncArg

class FuncParam:
    """A parameter from a function.

    This class is to create parameters directly from
    functions.

    Parameters
    ----------
        name : str
            Name of the parameter, by default
            the name of the function.

    Examples
    --------

    Simple example:

    >>> from rocketry.parameters import FuncParam
    >>> @FuncParam()
    ... def email_list():
    ...     ...
    ...     return ['me@example.com']

    Create a task that uses the parameter:

    >>> from rocketry.tasks import FuncTask
    >>> @FuncTask()
    >>> def send_things(email_list):
        ... # Send email list

    """
    def __init__(self, name=None, session=None):
        self.name = name
        self.session = session

    def __call__(self, func: Callable):
        session = FuncArg.session if self.session is None else self.session
        name = self._get_name(func)
        session.parameters[name] = FuncArg(func)
        func.__rocketry__ = {'param_name': name}
        return func

    def _get_name(self, func):
        if self.name is not None:
            return self.name
        return func.__name__





class CondParser:
    "Utility class for parsing a condition item"
    cache = {}
    def __init__(self, func, session, cached=False):
        self.func = func
        self.session = session
        self.cached = cached

    def __call__(self, s:str, *args, **kwargs):
        session = self.session
        if self.cached and s in session._cond_cache:
            return session._cond_cache[s]

        cond = self.func(*args, **kwargs)
        if self.cached:
            session._cond_cache[s] = cond
        return cond



import pytest
from rocketry.conditions.task.task import TaskStarted
from rocketry.core.condition.base import BaseCondition
from rocketry.tasks import FuncTask

from rocketry.conditions import FuncCond
from rocketry.conds import true

def run_succeeding():
    pass

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_func_cond_decorator(session, execution):

    @FuncCond(decor_return_func=False)
    def is_true():
        return True

    assert isinstance(is_true, BaseCondition)

    task_success = FuncTask(
        run_succeeding,
        start_cond=true & is_true,
        name="task success",
        execution=execution,
        session=session
    )

    session.config.shut_cond = TaskStarted(task=task_success) >= 1
    session.start()

    n_success = task_success.logger.filter_by(action="run").count()
    assert n_success == 1



import datetime
import logging
from queue import Empty
import multiprocessing

import pytest

from redbird.logging import RepoHandler

from rocketry.log.log_record import  MinimalRecord
from rocketry.tasks import FuncTask
from rocketry.testing.log import create_task_record

def run_success():
    pass

@pytest.mark.parametrize("optimized", [True, False])
@pytest.mark.parametrize("last_status", ["run", "success", "fail", "inaction", "terminate"])
def test_set_cached_in_init(session, optimized, last_status):
    session.config.force_status_from_logs = not optimized

    logger = logging.getLogger("rocketry.task")
    for handler in logger.handlers:
        if hasattr(handler, "repo"):
            repo = handler.repo

    times = {
        "run": 946677600.0,
        "success": 1656709200.0,
        "fail": 1656795600.0,
        "terminate": 1656882000.0,
        "inaction":1656968400.0,
    }


    # A log record that should not be used
    for action, created in times.items():
        repo.add(MinimalRecord(
            task_name="mytask",
            action=action,
            created=created
        ))
    # Add one extra (which should be used instead of what we did above)
    times[last_status] = 1752958800.0
    repo.add(MinimalRecord(
        task_name="mytask",
        action=last_status,
        created=1752958800.0
    ))

    # Creating the task
    task = FuncTask(
        lambda : None,
        execution="main",
        name="mytask",
        session=session
    )
    for action, created in times.items():
        dt = datetime.datetime.fromtimestamp(created)
        last_action_value = getattr(task, f"last_{action}")
        assert last_action_value == dt
    if last_status == "run":
        # The task was running according to logs when the task was created
        # Therefore it is considered crashed
        assert task.status == "crash"
    else:
        assert task.status == last_status

def test_running(session):

    task = FuncTask(
        lambda : None,
        execution="main",
        session=session
    )
    task.log_running()
    assert "run" == task.status
    assert task.is_running

def test_success(tmpdir, session):

    task = FuncTask(
        lambda : None,
        execution="main",
        session=session
    )
    task.log_running()
    task.log_success()
    assert "success" == task.status
    assert not task.is_running

def test_fail(tmpdir, session):

    task = FuncTask(
        lambda : None,
        execution="main",
        session=session
    )
    task.log_running()
    task.log_failure()
    assert "fail" == task.status
    assert not task.is_running

def test_without_running(tmpdir, session):
    "An edge case if for mysterious reason the task did not log running. Logging should still not crash"

    task = FuncTask(
        lambda : None,
        execution="main",
        session=session
    )
    task.log_failure()
    assert "fail" == task.status
    assert not task.is_running

def test_handle(session):

    def create_record(action, task_name):
        # Util func to create a LogRecord
        record = create_task_record(
            task_name=task_name,
            action=action,
            exc_info=None,
            # These should not matter:
            name="rocketry.task._process",
        )
        return record

    task = FuncTask(
        lambda : None,
        execution="main",
        name="a task",
        session=session
    )
    # Creating the run log
    record_run = create_record("run", task_name="a task")

    # Creating the outcome log
    record_finish = create_record("success", task_name="a task")

    task.log_record(record_run)
    assert "run" == task.status
    assert task.is_running

    task.log_record(record_finish)
    assert "success" == task.status
    assert not task.is_running

    records = session.get_task_log()
    records = [
        record.dict(exclude={"created"})
        for record in records
    ]
    assert [
        {"task_name": "a task", "action": "run"},
        {"task_name": "a task", "action": "success"},
    ] == records

def test_without_handlers(session):
    session.config.force_status_from_logs = True
    session.config.task_logger_basename = 'hdlr_test.task'

    logger = logging.getLogger("hdlr_test.task")
    logger.handlers = []
    logger.propagate = False
    logger.setLevel(logging.INFO)

    with pytest.warns(UserWarning) as warns:
        task = FuncTask(
            lambda : None,
            name="task 1",
            start_cond="always true",
            #logger="rocketry.task.test",
            execution="main",
            session=session
        )

    # Test warnings

    #assert str(warns[0].message) == "Logger 'rocketry.task.test' for task 'task 1' does not have ability to be read. Past history of the task cannot be utilized."
    warn_messages = [str(w.message) for w in warns]
    assert warn_messages == [
        "Logger hdlr_test.task cannot be read. Logging is set to memory. To supress this warning, please set a handler that can be read (redbird.logging.RepoHandler)"
    ]

    assert len(logger.handlers) == 1
    assert isinstance(logger.handlers[0], RepoHandler)

def test_without_handlers_status_warnings(session):
    session.config.force_status_from_logs = True

    logger = logging.getLogger("rocketry.task")
    logger.handlers = []
    logger.propagate = False
    logger.setLevel(logging.INFO)

    with pytest.warns(UserWarning) as warns:
        task = FuncTask(
            lambda : None,
            name="task 1",
            start_cond="always true",
            logger_name="rocketry.task.test",
            execution="main",
            session=session
        )
    # Removing the handlers that were added

    # Test warnings
    expected_warnings = [
        'Logger rocketry.task cannot be read. Logging is set to memory. To supress this warning, please set a handler that can be read (redbird.logging.RepoHandler)',
        "Logger 'rocketry.task.test' for task 'task 1' does not have ability to be read. Past history of the task cannot be utilized.",
        "Task 'task 1' logger is not readable. Latest run unknown.",
        "Task 'task 1' logger is not readable. Latest success unknown.",
        "Task 'task 1' logger is not readable. Latest fail unknown.",
        "Task 'task 1' logger is not readable. Latest terminate unknown.",
        "Task 'task 1' logger is not readable. Latest inaction unknown.",
        "Task 'task 1' logger is not readable. Latest crash unknown.",
    ]
    actual_warnings = [str(warn.message) for warn in warns]
    assert expected_warnings == actual_warnings

    # Removing the handlers that were added
    # to test without handlers
    logger.handlers = []

    with pytest.warns(UserWarning) as warns:
        task()
    # Cannot know the task.status as there is no log about it
    assert task.status == 'success'
    with pytest.warns(UserWarning) as warns:
        assert task.get_status() is None

    assert list(str(w.message) for w in warns) == [
        "Logger 'rocketry.task.test' for task 'task 1' does not have ability to be read. Past history of the task cannot be utilized.",
        "Task 'task 1' logger is not readable. Status unknown."
    ]

    session.config.force_status_from_logs = False
    with pytest.warns(UserWarning) as warns:
        task = FuncTask(
            lambda : None,
            name="task 2",
            start_cond="always true",
            logger="rocketry.task.test",
            execution="main",
            session=session
        )
    assert list(str(w.message) for w in warns) == [
        'Logger rocketry.task cannot be read. Logging is set to memory. To supress this warning, please set a handler that can be read (redbird.logging.RepoHandler)'
    ]

    task()
    # Can know the task.status as stored in a variable
    assert task.status == "success"
    assert task.get_status() == "success"

@pytest.mark.parametrize("method",
    [
        pytest.param("log_success", id="Success"),
        pytest.param("log_failure", id="Failure"),
        pytest.param("log_inaction", id="Inaction"),
        pytest.param("log_termination", id="Termination"),
    ],
)
def test_action_start(tmpdir, method, session):

    task = FuncTask(
        lambda : None,
        execution="main",
        session=session
    )
    task.log_running()
    getattr(task, method)()

    records = list(map(lambda e: e.dict(), session.get_task_log()))
    assert len(records) == 2

    # First should not have "end"
    first = records[0]
    assert first['created'] >= datetime.datetime(2000, 1, 1).timestamp()

    # Second should and that should be datetime
    last = records[1]
    assert first['created'] <= last['created']
    assert last['created'] < datetime.datetime(2200, 1, 1).timestamp()

def test_process_no_double_logging(session):
    # 2021-02-27 there is a bug that Raspbian logs process task logs twice
    # while this is not occuring on Windows. This tests the bug.
    #!NOTE: This test requires there are two handlers in
    # rocketry.task logger (Memory and Stream in this order)
    task_logger = logging.getLogger("rocketry.task")
    task_logger.addHandler(logging.StreamHandler())

    expected_actions = ["run", "success"]

    task = FuncTask(
        run_success,
        name="a task",
        execution="process",
        session=session
    )

    log_queue = multiprocessing.Queue(-1)

    # Start the process
    proc = multiprocessing.Process(target=task._run_as_process, args=(None, None, log_queue, session.config, []), daemon=None)
    proc.start()

    # Do the logging manually (copied from the method)
    actual_actions = []
    records = []
    while True:
        try:
            record = log_queue.get(block=True, timeout=3)
        except Empty:
            break
        else:
            records.append(record)
            # task.log_record(record)
            actual_actions.append(record.action)

    # If fails here, double logging caused by creating too many records
    # Obseved to fail rarely with py36 (c2f0368ffa56c5b8933f1afa2917f2be1555fb7a)
    assert len(records) >= 2 # Testing not too few records (else not double logging bug)
    assert (
        ["run", "success"] == [rec.action for rec in records]
    ), "Double logging. Caused by creating multiple records (Task.log_running & Task.log_success) to the queue."

    # Testing there is no log records yet in the log
    # (as the records should not been handled yet)
    recs = list(session.get_task_log())
    assert len(recs) == 0, "Double logging. The log file is not empty before handling the records. Process bypasses the queue."

    for record in records:
        task.log_record(record)

    # If fails here, double logging caused by multiple handlers
    handlers = task.logger.logger.handlers
    handlers[0] # Checking it has atleast 2
    handlers[1] # Checking it has atleast 2
    assert (
        isinstance(handlers[1], logging.StreamHandler)
        and isinstance(handlers[0], RepoHandler)
        and len(handlers) == 2
    ), f"Double logging. Too many handlers: {handlers}"

    # If fails here, double logging caused by Task.log_record
    n_run = len(list(session.get_task_log(task_name="a task", action="run")))
    n_success = len(list(session.get_task_log(task_name="a task", action="success")))

    assert n_run > 0 and n_success > 0, "No log records formed to log."

    assert n_run == 1 and n_success == 1, "Double logging. Bug in Task.log_record probably."



from rocketry.core.time import Any, All
from ..utils.string_parser import InstructionParser
from .time_item import parse_time_item

def _flatten(*args, types, with_attr):
    comps = []
    for arg in args:
        if isinstance(arg, types):
            comps += getattr(arg, with_attr)
        else:
            comps.append(arg)
    return comps

def _parse_any(*args):
    comps = _flatten(*args, types=Any, with_attr="periods")
    return Any(*comps)

def _parse_all(*args):
    comps = _flatten(*args, types=All, with_attr="periods")
    return All(*comps)

parse_time_string = InstructionParser(
    parse_time_item,
    operators=[
        {
            "symbol": "&",
            "func": _parse_all,
            "side": "both",
        },
        {
            "symbol": "|",
            "func": _parse_any,
            "side": "both",
        },
    ]
)



from rocketry.core.condition.base import BaseCondition
from rocketry.conds import true, false
from ._condition import parse_condition_string
from .utils import ParserPicker

def _parse_condition_string(s:str, **kwargs) -> BaseCondition:
    cond = parse_condition_string(s, **kwargs)
    cond._str = s
    return cond

def _parse_bool(b:bool) -> BaseCondition:
    return true if b else false

PARSER = ParserPicker(
    {
        str: _parse_condition_string,
        bool: _parse_bool
    }
)

def parse_condition(conf, **kwargs):
    if isinstance(conf, BaseCondition):
        return conf
    return PARSER(conf, **kwargs)




import pytest
from rocketry.pybox import query

@pytest.mark.parametrize('qry_dict,data,expected',
    [
        pytest.param(
            {
                'mykey$min': 10,
                'mykey$max': 20,
            },
            [{'mykey': 5}, {'mykey': 30}, {'mykey': 15}, {'mykey': 10}, {'mykey': 20}],
            [{'mykey': 15}, {'mykey': 10}, {'mykey': 20}],
            id='min-max'
        ),
        pytest.param(
            {
                'mykey': 10,
            },
            [{'mykey': 5}, {'mykey': 30}, {'mykey': 15}, {'mykey': 10}, {'mykey': 20}],
            [{'mykey': 10}],
            id='equal'
        ),
        pytest.param(
            {
                'mykey$not': 10,
            },
            [{'mykey': 5}, {'mykey': 30}, {'mykey': 15}, {'mykey': 10}, {'mykey': 20}],
            [{'mykey': 5}, {'mykey': 30}, {'mykey': 15}, {'mykey': 20}],
            id='not-equal'
        ),
        pytest.param(
            {
                'mykey$regex': '.+ or .+',
            },
            [{'mykey': 'foo or bar'}, {'mykey': 'bar or foo'}, {'mykey': ''}, {'mykey': 'foo and bar'}],
            [{'mykey': 'foo or bar'}, {'mykey': 'bar or foo'}],
            id='regex'
        ),
    ]
)
def test_from_dict_string(qry_dict, data, expected):
    qry = query.parser.from_dict(qry_dict)
    actual = qry.filter(data)
    assert list(actual) == expected


@pytest.mark.parametrize('qry_dict,data,expected',
    [
        pytest.param(
            [
                ('mykey$min', 10),
                ('mykey$max', 20),
            ],
            [{'mykey': 5}, {'mykey': 30}, {'mykey': 15}, {'mykey': 10}, {'mykey': 20}],
            [{'mykey': 15}, {'mykey': 10}, {'mykey': 20}],
            id='min-max'
        ),
        pytest.param(
            [
                ('mykey', 10),
            ],
            [{'mykey': 5}, {'mykey': 30}, {'mykey': 15}, {'mykey': 10}, {'mykey': 20}],
            [{'mykey': 10}],
            id='equal'
        ),
        pytest.param(
            [
                ('mykey$not', 10),
            ],
            [{'mykey': 5}, {'mykey': 30}, {'mykey': 15}, {'mykey': 10}, {'mykey': 20}],
            [{'mykey': 5}, {'mykey': 30}, {'mykey': 15}, {'mykey': 20}],
            id='not-equal'
        ),
        pytest.param(
            [
                ('mykey$regex', '.+ or .+'),
            ],
            [{'mykey': 'foo or bar'}, {'mykey': 'bar or foo'}, {'mykey': ''}, {'mykey': 'foo and bar'}],
            [{'mykey': 'foo or bar'}, {'mykey': 'bar or foo'}],
            id='regex'
        ),
        pytest.param(
            [
                ('mykey', 'bad'),
                ('mykey', 'good'),
            ],
            [{'mykey': 'bad'}, {'mykey': 'neutral'}, {'mykey': 'good'}],
            [{'mykey': 'bad'}, {'mykey': 'good'}],
            id='in'
        ),
        pytest.param(
            [
                ('mykey1', 'bad'),
                ('mykey1', 'good'),
                ('mykey2', 'happy'),
            ],
            [{'mykey1': 'bad', 'mykey2': 'sad'}, {'mykey1': 'neutral', 'mykey2': 'happy'}, {'mykey1': 'good', 'mykey2': 'happy'}],
            [{'mykey1': 'good', 'mykey2': 'happy'}],
            id='combined'
        ),
    ]
)
def test_from_tuples(qry_dict, data, expected):
    qry = query.parser.from_tuples(qry_dict)
    actual = qry.filter(data)
    assert list(actual) == expected


@pytest.mark.parametrize('qry_kwargs,data,expected',
    [
        pytest.param(
            {'mykey': (10, 15)},
            [{'mykey': 5}, {'mykey': 10}, {'mykey': 12}, {'mykey': 15}, {'mykey': 30}],
            [{'mykey': 10}, {'mykey': 12}, {'mykey': 15}],
            id='min-max'
        ),
        pytest.param(
            {'mykey': (10, None)},
            [{'mykey': 5}, {'mykey': 10}, {'mykey': 12}, {'mykey': 15}, {'mykey': 30}],
            [{'mykey': 10}, {'mykey': 12}, {'mykey': 15}, {'mykey': 30}],
            id='min'
        ),
        pytest.param(
            {'mykey': (None, 15)},
            [{'mykey': 5}, {'mykey': 10}, {'mykey': 12}, {'mykey': 15}, {'mykey': 30}],
            [{'mykey': 5}, {'mykey': 10}, {'mykey': 12}, {'mykey': 15}],
            id='max'
        ),
        pytest.param(
            {'mykey': ['good', 'bad']},
            [{'mykey': 'good'}, {'mykey': 'bad'}, {'mykey': 'ugly'}],
            [{'mykey': 'good'}, {'mykey': 'bad'}],
            id='in'
        ),
    ]
)
def test_from_kwargs(qry_kwargs, data, expected):
    qry = query.parser.from_kwargs(**qry_kwargs)
    actual = qry.filter(data)
    assert list(actual) == expected



import asyncio
import datetime
import logging
import time
import os
import re
import multiprocessing

import pytest

from redbird.logging import RepoHandler
from redbird.repos import MemoryRepo

from rocketry.log.log_record import TaskLogRecord
import rocketry
from rocketry import Session
from rocketry.core import Parameters
from rocketry.log.log_record import MinimalRecord
from rocketry.tasks import FuncTask
from rocketry.time import TimeDelta
from rocketry.exc import TaskInactionException, TaskTerminationException
from rocketry.conditions import SchedulerCycles, SchedulerStarted, TaskStarted, AlwaysFalse, AlwaysTrue
from rocketry.args import Private, TerminationFlag

from rocketry.conds import true, false

def run_failing():
    raise RuntimeError("Task failed")

def run_succeeding():
    pass

def run_inacting():
    raise TaskInactionException()

async def run_failing_async():
    raise RuntimeError("Task failed")

async def run_succeeding_async():
    pass

async def run_inacting_async():
    raise TaskInactionException()


def run_slow():
    time.sleep(5)

def run_slow_thread(flag=TerminationFlag()):
    t = 0
    while not flag.is_set() and t < 5:
        time.sleep(0.001)
        t += 0.001
    if flag.is_set():
        raise TaskTerminationException()

async def run_slow_async():
    await asyncio.sleep(5)


def create_line_to_file():
    with open("work.txt", "a") as file:
        file.write("line created\n")

async def create_line_to_file_async():
    with open("work.txt", "a") as file:
        file.write("line created\n")

def run_with_param(int_5):
    assert int_5 == 5

def run_creating_child():

    proc = multiprocessing.Process(target=run_succeeding, daemon=True)
    proc.start()

def test_scheduler_shut_cond(session):
    assert not session.scheduler.check_shut_cond(None)

    assert session.scheduler.check_shut_cond(true)
    assert not session.scheduler.check_shut_cond(~true)
    assert session.scheduler.check_shut_cond(true & true)
    assert not session.scheduler.check_shut_cond(false)

@pytest.mark.parametrize("execution", ["main", "async", "thread", "process"])
@pytest.mark.parametrize("func", [pytest.param(create_line_to_file, id="sync"), pytest.param(create_line_to_file_async, id="async")])
def test_task_execution(tmpdir, execution, func, session):
    with tmpdir.as_cwd() as old_dir:
        # To be confident the scheduler won't lie to us
        # we test the task execution with a job that has
        # actual measurable impact outside rocketry
        FuncTask(func, name="add line to file", start_cond=AlwaysTrue(), execution=execution, session=session),

        session.config.shut_cond = (TaskStarted(task="add line to file") >= 3) | ~SchedulerStarted(period=TimeDelta("5 second"))

        session.start()
        # Sometimes in CI the task may end up to be started only twice thus we tolerate slightly
        with open("work.txt", "r") as file:
            assert 2 <= len(list(file))

@pytest.mark.parametrize("get_handler", [
    pytest.param(lambda: RepoHandler(repo=MemoryRepo(model=TaskLogRecord)), id="Memory with model"),
    pytest.param(lambda: RepoHandler(repo=MemoryRepo()), id="Memory with dict"),
])
@pytest.mark.parametrize("execution", ["main", "thread", "process"])
@pytest.mark.parametrize(
    "task_func,run_count,fail_count,success_count,inact_count",
    [
        pytest.param(
            run_succeeding,
            3, 0, 3, 0,
            id="Succeeding task"),

        pytest.param(
            run_failing,
            3, 3, 0, 0,
            id="Failing task"),
        pytest.param(
            run_inacting,
            3, 0, 0, 3,
            id="Inacting task"),
    ],
)
def test_task_log(tmpdir, execution, task_func, run_count, fail_count, success_count, inact_count, get_handler):
    """Test the task logging thoroughly including the common
    logging schemes, execution (main, thread, process) and
    outcomes (fail, success, inaction).

    This task may take some time but should cover most of the
    issues with logging.
    """

    # Set session (and logging)
    session = Session(config={"debug": True, "silence_task_logging": False, "task_execution": "process"})
    rocketry.session = session
    session.set_as_default()

    task_logger = logging.getLogger(session.config.task_logger_basename)
    task_logger.handlers = [
        get_handler()
    ]

    task = FuncTask(task_func, name="mytask", start_cond=AlwaysTrue(), execution=execution, session=session)

    session.config.shut_cond = (TaskStarted(task="mytask") >= run_count) | ~SchedulerStarted(period=TimeDelta("10 second"))
    session.start()

    assert (TaskStarted(task="mytask") >= run_count).observe(session=session)

    # Test history
    history = list(task.logger.get_records())
    logger = task.logger
    assert run_count == logger.filter_by(action="run").count()
    assert success_count == logger.filter_by(action="success").count()
    assert fail_count == logger.filter_by(action="fail").count()
    assert inact_count == logger.filter_by(action="inaction").count()

    # Test relevant log items
    for record in history:
        if not isinstance(record, dict):
            record = record.dict()
        assert record["task_name"] == "mytask"
        assert isinstance(record["created"], float)
        assert isinstance(record["start"], datetime.datetime)
        if record["action"] != "run":
            assert isinstance(record["end"], datetime.datetime)
            assert isinstance(record["runtime"], datetime.timedelta)

        # Test traceback
        if record["action"] == "fail":
            assert re.match(
                r'Traceback \(most recent call last\):\n  File ".+", line [0-9]+, in [\s\S]+, in run_failing\n    raise RuntimeError\("Task failed"\)\nRuntimeError: Task failed',
                record["exc_text"]
            )

    # Test some other relevant APIs
    assert history == list(task.logger.get_records())
    assert history[-1] == task.logger.get_latest()
    assert history == list(session.get_task_log())

    assert run_count == len(list(task.logger.get_records(action="run")))
    assert success_count == len(list(task.logger.get_records(action="success")))
    assert fail_count == len(list(task.logger.get_records(action="fail")))
    assert inact_count == len(list(task.logger.get_records(action="inaction")))

@pytest.mark.parametrize("mode", ["use logs", "use cache"])
@pytest.mark.parametrize("func_type", ["sync", "async"])
@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_task_status(session, execution, func_type, mode):
    session.config.force_status_from_logs = True if mode == "use logs" else False

    task_success = FuncTask(
        run_succeeding if func_type == "sync" else run_succeeding_async,
        start_cond=TaskStarted(task="task success") < 3,
        name="task success",
        execution=execution,
        priority=0,
        session=session
    )
    task_fail = FuncTask(
        run_failing if func_type == "sync" else run_failing_async,
        start_cond=TaskStarted(task="task fail") < 3,
        name="task fail",
        execution=execution,
        priority=0,
        session=session
    )
    task_inact = FuncTask(
        run_inacting if func_type == "sync" else run_inacting_async,
        start_cond=TaskStarted(task="task inact") < 3,
        name="task inact",
        execution=execution,
        priority=100,
        session=session
    )
    task_not_run = FuncTask(
        run_inacting if func_type == "sync" else run_inacting_async,
        start_cond=false,
        name="task not run",
        execution=execution,
        priority=0,
        session=session
    )
    session.config.shut_cond = (
        (TaskStarted(task="task success") >= 3)
        & (TaskStarted(task="task fail") >= 3)
        & (TaskStarted(task="task inact") >= 3)
    ) | ~SchedulerStarted(period=TimeDelta("20 second"))
    session.start()

    # Test status
    assert task_success.last_run is not None
    assert task_success.last_success is not None
    assert task_success.last_fail is None

    assert task_fail.last_run is not None
    assert task_fail.last_fail is not None
    assert task_fail.last_success is None

    assert task_inact.last_run is not None

    assert task_not_run.last_run is None
    assert task_not_run.last_success is None
    assert task_not_run.last_fail is None

    assert task_success.status == "success"
    assert task_fail.status == "fail"
    assert task_inact.status == "inaction"
    assert task_not_run.status is None

    # Test logs
    assert 3 == task_success.logger.filter_by(action="run").count()
    assert 3 == task_fail.logger.filter_by(action="run").count()
    assert 3 == task_inact.logger.filter_by(action="run").count()
    assert 0 == task_not_run.logger.filter_by(action="run").count()

    assert 3 == task_success.logger.filter_by(action="success").count()
    assert 0 == task_fail.logger.filter_by(action="success").count()
    assert 0 == task_inact.logger.filter_by(action="suceess").count()

    assert 0 == task_success.logger.filter_by(action="fail").count()
    assert 3 == task_fail.logger.filter_by(action="fail").count()
    assert 0 == task_inact.logger.filter_by(action="fail").count()

    assert 0 == task_success.logger.filter_by(action="inaction").count()
    assert 0 == task_fail.logger.filter_by(action="inaction").count()
    assert 3 == task_inact.logger.filter_by(action="inaction").count()




@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_task_disabled(tmpdir, execution, session):
    with tmpdir.as_cwd() as old_dir:

        task = FuncTask(
            run_succeeding,
            start_cond=AlwaysFalse(),
            name="task",
            execution=execution,
            session=session
        )
        task.disabled = True

        session.config.shut_cond = SchedulerCycles() >= 5
        session.start()

        history = task.logger.get_records()
        assert 0 == sum([record for record in history if record["action"] == "run"])

        assert task.disabled

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_priority(execution, session):
    session.config.max_process_count = 4
    task_1 = FuncTask(run_succeeding, name="1", priority=100, start_cond=AlwaysTrue(), execution=execution, session=session)
    task_3 = FuncTask(run_failing, name="3", priority=10, start_cond=AlwaysTrue(), execution=execution, session=session)
    task_2 = FuncTask(run_failing, name="2", priority=50, start_cond=AlwaysTrue(), execution=execution, session=session)
    task_4 = FuncTask(run_failing, name="4", start_cond=AlwaysTrue(), execution=execution, session=session)

    assert 0 == task_4.priority

    session.config.shut_cond = (SchedulerCycles() == 1) | ~SchedulerStarted(period=TimeDelta("20 seconds"))

    session.start()
    assert session.scheduler.n_cycles == 1

    task_1_start = list(task_1.logger.get_records())[0].created
    task_2_start = list(task_2.logger.get_records())[0].created
    task_3_start = list(task_3.logger.get_records())[0].created
    task_4_start = list(task_4.logger.get_records())[0].created

    assert task_1_start < task_2_start < task_3_start < task_4_start

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_pass_params_as_global(execution, session):
    # thread-Parameters has been observed to fail rarely

    task = FuncTask(run_with_param, name="parametrized", start_cond=AlwaysTrue(), execution=execution, session=session)

    session.config.shut_cond = (TaskStarted(task="parametrized") >= 1) | ~SchedulerStarted(period=TimeDelta("20 seconds"))

    # Passing global parameters
    session.parameters["int_5"] = 5
    session.parameters["extra_param"] = "something"

    session.scheduler()

    logger = task.logger
    assert 1 == logger.filter_by(action="run").count()
    assert 1 == logger.filter_by(action="success").count()
    assert 0 == logger.filter_by(action="fail").count()

@pytest.mark.parametrize("parameters", [
    pytest.param({"int_5": 5}, id="dict"),
    pytest.param(Parameters(int_5=5), id="Parameters"),
    pytest.param(Parameters(int_5=Private(5)), id="Parameter with secret"),
])
@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_pass_params_as_local(execution, parameters, session):

    task = FuncTask(
        run_with_param,
        name="parametrized",
        parameters=parameters,
        start_cond=AlwaysTrue(),
        execution=execution,
        session=session
    )
    session.config.shut_cond = (TaskStarted(task="parametrized") >= 1) | ~SchedulerStarted(period=TimeDelta("20 seconds"))

    session.start()

    logger = task.logger
    assert 1 == logger.filter_by(action="run").count()
    assert 1 == logger.filter_by(action="success").count()
    assert 0 == logger.filter_by(action="fail").count()

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_pass_params_as_local_and_global(execution, session):

    task = FuncTask(
        run_with_param,
        name="parametrized",
        parameters={"int_5": 5},
        start_cond=AlwaysTrue(),
        execution=execution,
        session=session
    )

    session.config.shut_cond = (TaskStarted(task="parametrized") >= 1) | ~SchedulerStarted(period=TimeDelta("20 seconds"))

    # Additional parameters
    session.parameters["extra_param"] = "something"

    session.start()

    logger = task.logger
    assert 1 == logger.filter_by(action="run").count()
    assert 1 == logger.filter_by(action="success").count()
    assert 0 == logger.filter_by(action="fail").count()


# Only needed for testing start up and shutdown
def create_line_to_startup_file():
    with open("start.txt", "w") as file:
        file.write("line created\n")

def create_line_to_shutdown():
    with open("shut.txt", "w") as file:
        file.write("line created\n")

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_startup_shutdown(tmpdir, execution, session):
    with tmpdir.as_cwd() as old_dir:

        FuncTask(create_line_to_startup_file, name="startup", on_startup=True, execution=execution, session=session)
        FuncTask(create_line_to_shutdown, name="shutdown", on_shutdown=True, execution=execution, session=session)

        session.config.shut_cond = AlwaysTrue()

        session.start()
        if execution == "thread":
            # It may take a moment for the
            # thread to write to the file
            timeout = 5
            time_taken = 0
            while os.path.exists("shut.txt"):
                time.sleep(0.01)
                time_taken += 0.01
                if time_taken > timeout:
                    break

        assert os.path.exists("start.txt")
        assert os.path.exists("shut.txt")

        assert list(session.get_task_log()) != []

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_logging_repo(tmpdir, execution):
    session = Session(config={'task_execution': 'async'})
    session.set_as_default()
    session.config.max_process_count = 4

    handler = RepoHandler(repo=MemoryRepo(model=MinimalRecord))

    logger = logging.getLogger("rocketry.task")
    logger.handlers = []
    logger.addHandler(handler)


    with tmpdir.as_cwd() as old_dir:

        task_1 = FuncTask(run_succeeding, name="1", priority=100, start_cond=AlwaysTrue(), execution=execution, session=session)
        task_3 = FuncTask(run_failing, name="3", priority=10, start_cond=AlwaysTrue(), execution=execution, session=session)
        task_2 = FuncTask(run_failing, name="2", priority=50, start_cond=AlwaysTrue(), execution=execution, session=session)
        task_4 = FuncTask(run_failing, name="4", start_cond=AlwaysTrue(), execution=execution, session=session)

        assert 0 == task_4.priority

        session.config.shut_cond = (SchedulerCycles() == 1) | ~SchedulerStarted(period=TimeDelta("20 seconds"))
        session.start()
        assert session.scheduler.n_cycles == 1

        task_1_start = list(task_1.logger.get_records())[0].created
        task_2_start = list(task_2.logger.get_records())[0].created
        task_3_start = list(task_3.logger.get_records())[0].created
        task_4_start = list(task_4.logger.get_records())[0].created

        assert task_1_start < task_2_start < task_3_start < task_4_start

@pytest.mark.parametrize("execution", ["async", "thread", "process"])
def test_instant_shutdown(execution, session):
    assert not session.config.instant_shutdown
    session.config.instant_shutdown = True

    func = {"async": run_slow_async, "thread": run_slow_thread, "process": run_slow}[execution]

    task = FuncTask(func, execution=execution, start_cond=true, end_cond=SchedulerCycles() == 2, session=session)

    session.config.shut_cond = SchedulerCycles() == 2

    session.start()

    assert 1 == task.logger.filter_by(action="run").count()
    assert 2 == task.logger.filter_by().count()

    assert 0 == task.logger.filter_by(action="fail").count()
    assert 0 == task.logger.filter_by(action="success").count()
    assert 1 == task.logger.filter_by(action="terminate").count()

def test_cycle_sleep_none(session):
    assert not session.config.instant_shutdown
    session.config.instant_shutdown = True
    session.config.cycle_sleep = None

    order = []

    async def run_async():
        order.append("async start")
        await asyncio.sleep(0)
        order.append("async end")

    def run_shutdown():
        order.append("shutdown task")

    FuncTask(run_async, execution="async", start_cond=true, session=session)
    FuncTask(run_shutdown, execution="main", start_cond=true, session=session)

    session.config.shut_cond = SchedulerCycles() == 1

    session.start()

    assert order == [
        "async start",
        "shutdown task",
        "async end"
    ]



from typing import Iterable, Iterator
import datetime

from rocketry.pybox.time.convert import to_datetime, to_timedelta

class QueryBase:
    
    def _get_value(self, item:dict, oper):
        return item[oper.name] if isinstance(oper, Key) else oper

    def filter(self, data:Iterable[dict]) -> Iterator[dict]:
        "Filter iterable of dicts"
        for dict in data:
            if self.match(dict):
                yield dict

    def to_pykwargs(self):
        """Put the query to simple Python keyword representation"""

        def _get_key_value(left, right):
            is_left_key = isinstance(left, Key)
            is_right_key = isinstance(right, Key)
            if is_left_key and is_right_key:
                raise ValueError("Both cannot be keys")
            elif is_left_key:
                return left.name, right
            elif is_right_key:
                return right.name, left
            else:
                raise ValueError("Neither are keys")
        py_query = {}
        qry = self
        if self == true:
            return {}
        if not isinstance(self, All):
            qry = (self,)
        for sub_qry in qry:
            if isinstance(sub_qry, Equal):
                key, value = _get_key_value(sub_qry.left, sub_qry.right)
                py_query[key] = value
            elif isinstance(sub_qry, (GreaterEqual, LessEqual)):
                key, value = _get_key_value(sub_qry.left, sub_qry.right)
                past_min_val = py_query.get(key, (None, None))[0]
                past_max_val = py_query.get(key, (None, None))[1]
                if isinstance(sub_qry, LessEqual):
                    min_val = past_min_val
                    max_val = value
                elif isinstance(sub_qry, GreaterEqual):
                    min_val = value
                    max_val = past_max_val
                py_query[key] = (min_val, max_val)
            else:
                raise TypeError(f"No conversion for type: {type(sub_qry)}")
        return py_query


class Key(QueryBase):
    """Key of a data dictionary"""
    def __init__(self, name):
        self.name = name

    def __eq__(self, other):
        return Equal(self, other)

    def __ne__(self, other):
        return NotEqual(self, other)

    def __gt__(self, other):
        return Greater(self, other)

    def __lt__(self, other):
        return Less(self, other)

    def __ge__(self, other):
        return GreaterEqual(self, other)

    def __le__(self, other):
        return LessEqual(self, other)

    def __repr__(self):
        return f'Key({repr(self.name)})'

    def __str__(self):
        return f'<{self.name}>'

    def get_value(self, d:dict):
        return d[self.name]

class Expression(QueryBase):
    """Base class for query expressions"""
    def __and__(self, other):
        if not isinstance(self, All):
            self = (self,)
        if not isinstance(other, All):
            other = (other,)
        return All(*self, *other)

    def __or__(self, other):
        if not isinstance(self, Any):
            self = (self,)
        if not isinstance(other, Any):
            other = (other,)
        return Any(*self, *other)

    def __invert__(self):
        return Not(self)

    def _to_comparable(self, left, right):
        dt_cls = (datetime.datetime,)
        is_datetime = isinstance(left, dt_cls) or isinstance(right, dt_cls)
        if is_datetime:
            return to_datetime(left), to_datetime(right)
        else:
            return left, right

class All(Expression):

    def __init__(self, *args):
        self.args = args

    def match(self, item:dict):
        return all(arg.match(item) for arg in self.args)

    def __iter__(self):
        return iter(self.args)

    def __repr__(self):
        str_args = ', '.join(repr(arg) for arg in self.args)
        return f'{type(self).__name__}({str_args})'

    def __str__(self):
        return '(' + ' & '.join(str(arg) for arg in self.args) + ')'

class Any(Expression):

    def __init__(self, *args):
        self.args = args

    def match(self, item:dict):
        return any(arg.match(item) for arg in self.args)

    def __iter__(self):
        return iter(self.args)

    def __str__(self):
        return '(' + ' | '.join(str(arg) for arg in self.args) + ')'

class Not(Expression):

    def __init__(self, right):
        self.right = right

    def match(self, item:dict):
        return not self.right.match(item)

    def __str__(self):
        return f'~{self.right}'


class Equal(Expression):

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def match(self, item:dict):
        left_value = self._get_value(item, oper=self.left)
        right_value = self._get_value(item, oper=self.right)
        left_value, right_value = self._to_comparable(left_value, right_value)
        return left_value == right_value

    def __str__(self):
        left, right = self.left, self.right
        left = repr(left) if isinstance(left, str) else str(left)
        right = repr(right) if isinstance(right, str) else str(right)
        return '(' + f"{str(left)} == {str(right)}" + ')'

class NotEqual(Expression):

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def match(self, item:dict):
        left_value = self._get_value(item, oper=self.left)
        right_value = self._get_value(item, oper=self.right)
        left_value, right_value = self._to_comparable(left_value, right_value)
        return left_value != right_value

    def __str__(self):
        left, right = self.left, self.right
        left = repr(left) if isinstance(left, str) else str(left)
        right = repr(right) if isinstance(right, str) else str(right)
        return '(' + f"{str(left)} != {str(right)}" + ')'

class Greater(Expression):

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def match(self, item:dict):
        left_value = self._get_value(item, oper=self.left)
        right_value = self._get_value(item, oper=self.right)
        left_value, right_value = self._to_comparable(left_value, right_value)
        return left_value > right_value

    def __str__(self):
        left, right = self.left, self.right
        left = repr(left) if isinstance(left, str) else str(left)
        right = repr(right) if isinstance(right, str) else str(right)
        return '(' + f"{str(left)} > {str(right)}" + ')'

class GreaterEqual(Expression):

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def match(self, item:dict):
        left_value = self._get_value(item, oper=self.left)
        right_value = self._get_value(item, oper=self.right)
        left_value, right_value = self._to_comparable(left_value, right_value)
        return left_value >= right_value

    def __str__(self):
        left, right = self.left, self.right
        left = repr(left) if isinstance(left, str) else str(left)
        right = repr(right) if isinstance(right, str) else str(right)
        return '(' + f"{str(left)} >= {str(right)}" + ')'

class Less(Expression):

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def match(self, item:dict):
        left_value = self._get_value(item, oper=self.left)
        right_value = self._get_value(item, oper=self.right)
        left_value, right_value = self._to_comparable(left_value, right_value)
        return left_value < right_value

    def __str__(self):
        left, right = self.left, self.right
        left = repr(left) if isinstance(left, str) else str(left)
        right = repr(right) if isinstance(right, str) else str(right)
        return '(' + f"{str(left)} < {str(right)}" + ')'

class LessEqual(Expression):

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def match(self, item:dict):
        left_value = self._get_value(item, oper=self.left)
        right_value = self._get_value(item, oper=self.right)
        left_value, right_value = self._to_comparable(left_value, right_value)
        return left_value <= right_value

    def __str__(self):
        left, right = self.left, self.right
        left = repr(left) if isinstance(left, str) else str(left)
        right = repr(right) if isinstance(right, str) else str(right)
        return '(' + f"{str(left)} <= {str(right)}" + ')'


class Boolean(Expression):

    def __init__(self, value):
        self.value = value

    def match(self, item:dict):
        return bool(self.value)

true = Boolean(True)
false = Boolean(False)



import pytest
from rocketry.args.builtin import SimpleArg
from rocketry.conditions.task.task import TaskStarted
from rocketry.core.parameters.parameters import Parameters
from rocketry.tasks import FuncTask
from rocketry.conditions import SchedulerCycles, AlwaysFalse

def run_succeeding():
    pass

def run_parametrized(arg=SimpleArg("incorrect")):
    assert arg == "correct"

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_set_running(execution, session):
    task = FuncTask(
        run_succeeding,
        start_cond=AlwaysFalse(),
        name="task",
        execution=execution,
        session=session
    )
    assert task.batches == []
    task.run()
    assert task.batches == [Parameters()]

    session.config.shut_cond = SchedulerCycles() >= 5
    session.start()

    logger = task.logger
    assert 1 == logger.filter_by(action="run").count()
    assert 1 == logger.filter_by(action="success").count()
    assert 0 == logger.filter_by(action="fail").count()

    assert len(task.batches) == 0

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_set_running_with_params(execution, session):
    task = FuncTask(
        run_parametrized,
        start_cond=AlwaysFalse(),
        name="task",
        execution=execution,
        session=session
    )
    task.run(arg="correct")
    task.run(arg="correct")
    task.run(arg="incorrect")
    assert task.batches == [
        Parameters({"arg": "correct"}), Parameters({"arg": "correct"}), Parameters({"arg": "incorrect"})
    ]

    session.config.shut_cond = TaskStarted(task=task) == 3
    session.start()

    logger = task.logger
    assert 3 == logger.filter_by(action="run").count()
    assert 2 == logger.filter_by(action="success").count()
    assert 1 == logger.filter_by(action="fail").count()

    assert len(task.batches) == 0

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_set_running_disabled(execution, session):
    # NOTE: force_run overrides disabled
    # as it is more practical to keep
    # a task disabled and force it running
    # manually than prevent force run with
    # disabling
    task = FuncTask(
        run_succeeding,
        start_cond=AlwaysFalse(),
        name="task",
        execution=execution,
        session=session
    )
    task.disabled = True
    task.run()

    session.config.shut_cond = SchedulerCycles() >= 5
    session.start()
    assert task.batches == []

    logger = task.logger
    assert 1 == logger.filter_by(action="run").count()
    assert 1 == logger.filter_by(action="success").count()

    assert task.disabled

# Deprecated
# ----------

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_task_force_run(execution, session):
    task = FuncTask(
        run_succeeding,
        start_cond=AlwaysFalse(),
        name="task",
        execution=execution,
        session=session
    )
    with pytest.warns(DeprecationWarning):
        task.force_run = True

    session.config.shut_cond = SchedulerCycles() >= 5
    session.start()

    logger = task.logger
    assert 1 == logger.filter_by(action="run").count()

    # The force_run should have reseted as it should have run once
    assert not task.force_run

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_task_force_disabled(execution, session):
    # NOTE: force_run overrides disabled
    # as it is more practical to keep
    # a task disabled and force it running
    # manually than prevent force run with
    # disabling
    task = FuncTask(
        run_succeeding,
        start_cond=AlwaysFalse(),
        name="task",
        execution=execution,
        session=session
    )
    task.disabled = True
    with pytest.warns(DeprecationWarning):
        task.force_run = True

    session.config.shut_cond = SchedulerCycles() >= 5
    session.start()
    assert task.batches == []

    logger = task.logger
    assert 1 == logger.filter_by(action="run").count()
    assert 1 == logger.filter_by(action="success").count()

    assert task.disabled
    assert not task.force_run # This should be reseted



import logging
import pytest
from rocketry import Rocketry

def test_set_logging():
    app = Rocketry(execution="async")
    with pytest.warns(DeprecationWarning):
        @app.set_logger()
        def set_logging(logger):
            assert isinstance(logger, logging.Logger)



import datetime
from typing import Union
from dateutil.parser import parse

ABBREVIATIONS = {
    'ns': 'nanosecond',
    'μs': 'microsecond',
    'ms': 'millisecond',
    's': 'second',
    'm': 'minute',
    'h': 'hour',

    'nanosecond': 'nanosecond',
    'microsecond': 'microsecond',
    'millisecond': 'millisecond',
    'second': 'second',
    'minute': 'minute',
    'hour': 'hour',
}

def datetime_to_dict(dt):
    return {
        "year": dt.year,
        "month": dt.month,
        "day": dt.day,
        "hour": dt.hour,
        "minute": dt.minute,
        "second": dt.second,
        "microsecond": dt.microsecond,
        "nanosecond": dt.nanosecond if hasattr(dt, "nanosecond") else 0
    }

def to_timestamp(dt) -> float:
    if hasattr(dt, "to_pydatetime"):
        dt = dt.to_pydatetime()
    return dt.timestamp()

def to_datetime(s):
    if isinstance(s, datetime.datetime):
        return s
    elif isinstance(s, str):
        return string_to_datetime(s)
    elif hasattr(s, "timestamp"):
        # Is datetime-like. Tests' monkeypatching
        # overrides datetime.datetime thus we cannot
        # always rely on type
        return datetime.datetime.fromtimestamp(s.timestamp())
    else:
        raise TypeError(f"Cannot convert to datetime: {type(s)}")

def to_timedelta(s, **kwargs):
    "Convert object to timedelta"
    if isinstance(s, datetime.timedelta):
        return s
    elif isinstance(s, str):
        return string_to_timedelta(s)
    elif isinstance(s, (int, float)):
        return numb_to_timedelta(s, **kwargs)
    else:
        raise TypeError(f"Cannot convert to timedelta: {type(s)}")

def timedelta_to_dict(dt, days_in_year=365, days_in_month=30, units=None):

    total_seconds = dt.total_seconds()

    microsec_in_day = 8.64e+10
    microsec_in_hour = 3.6e+9
    microsec_in_min = 6e+7
    microsec_in_sec = 1e+6

    microsec = total_seconds * microsec_in_sec

    components = {}
    if units == "all":
        allowed_units = (
            "years", "months", "weeks",
            "days", "hours", "minutes",
            "seconds", "milliseconds",
            "microseconds")
    elif units == "fixed" or units is None:
        # Only fixed length units
        allowed_units = (
            "days", "hours", "minutes",
            "seconds", "milliseconds",
            "microseconds")
    else:
        allowed_units = units

    # Non fixed units
    if "years" in allowed_units:
        components["years"] = int(microsec // (microsec_in_day * days_in_year))
        microsec = microsec - (components["years"] * microsec_in_day * days_in_year)

    if "months" in allowed_units:
        components["months"] = int(microsec // (microsec_in_day * days_in_month))
        microsec = microsec - (components["months"] * microsec_in_day * days_in_month)

    # Alternating units
    if "weeks" in allowed_units:
        components["weeks"] = int(microsec // (microsec_in_day * 7))
        microsec = microsec - components["weeks"] * microsec_in_day * 7

    # Fixed units
    if "days" in allowed_units:
        components["days"] = int(microsec // microsec_in_day)
        microsec = microsec - components["days"] * microsec_in_day

    if "hours" in allowed_units:
        components["hours"] = int(microsec // microsec_in_hour)
        microsec = microsec - components["hours"] * microsec_in_hour

    if "minutes" in allowed_units:
        components["minutes"] = int(microsec // microsec_in_min)
        microsec = microsec - components["minutes"] * microsec_in_min

    if "seconds" in allowed_units:
        components["seconds"] = int(microsec // microsec_in_sec)
        microsec = microsec - components["seconds"] * microsec_in_sec

    if "milliseconds" in allowed_units:
        components["milliseconds"] = int(microsec // 1_000)
        microsec = microsec - components["milliseconds"] * 1_000

    if "microseconds" in allowed_units:
        components["microseconds"] = int(microsec)
        microsec = microsec - components["microseconds"]

    return components

def timedelta_to_str(dt,
                     days_in_year=365, days_in_month=30,
                     sep=", ", format=None, include=None,
                     default_scope="microseconds"):


    components = timedelta_to_dict(dt, days_in_year=days_in_year, days_in_month=days_in_month, units=include)

    # Min and max units (components must be ordered from biggest to smallest units)
    min_unit = default_scope
    for unit, value in components.items():
        if value != 0:
            min_unit = unit

    max_unit = default_scope
    for unit, value in components.items():
        if value != 0:
            # Max currently found non zero
            max_unit = unit
            break

    # Determine units
    if format is None:
        format = {
            "years": " years",
            "months": " months",
            "weeks": " weeks",
            "days": " days",
            "hours": " hours",
            "minutes": " minutes",
            "seconds": " seconds",
            "milliseconds": " milliseconds",
            "microseconds": " microseconds",
        }
    elif format == "semishort":
        format = {
            "years": " years",
            "months": " months",
            "weeks": " weeks",
            "days": " days",
            "hours": " hrs",
            "minutes": " mins",
            "seconds": " secs",
            "milliseconds": " ms",
            "microseconds": " μs",
        }
    elif format in ("short", "abbrs"):
        format = {
            "years": "Y",
            "months": "M",
            "weeks": "W",
            "days": "d",
            "hours": "h",
            "minutes": "m",
            "seconds": "s",
            "milliseconds": "ms",
            "microseconds": "μs",
        }

    # String format
    s = ""
    is_min_reached = False
    is_max_reached = False
    for unit, value in components.items():
        is_min_reached = min_unit == unit
        is_max_reached = is_max_reached or unit == max_unit

        unit_name = format[unit]
        if max_unit == unit:
            s = s + f"{value}{unit_name}"
        elif is_max_reached:
            s = s + f"{sep}{value}{unit_name}"

        if is_min_reached:
            break
    return s

def string_to_datetime(s):
    return parse(s)


def numb_to_timedelta(n: Union[float, int], unit="μs"):

    if unit == "ns":
        unit = 'μs'
        n = n / 1000
    units = ABBREVIATIONS[unit] + "s"
    return datetime.timedelta(**{units: n})

def string_to_timedelta(s:str):
    "Convert string to timedelta"

    def is_numeric_char(s):
        return s.isdigit() or s in ('.',)

    def is_wordbreak_char(s):
        return s in (' ', ',', ':', ',')

    def skip_wordbreak(s):
        for pos, char in enumerate(s):
            if not is_wordbreak_char(char):
                break
        return pos

    def get_number(s):
        num = ""
        for pos, char in enumerate(s):
            if is_numeric_char(char):
                num += char
            else:
                break
        return num, pos

    def get_unit(s):
        abbr = ""
        for pos, char in enumerate(s):
            if is_numeric_char(char) or is_wordbreak_char(char):
                break
            else:
                abbr += char
        else:
            # String ended
            pos += 1
        return abbrs[abbr], pos

    def get_hhmmss(s):
        hh, mm, ss = s.split(":")
        return to_microseconds(hour=int(hh), minute=int(mm), second=float(ss))

    # https://github.com/pandas-dev/pandas/blob/e8093ba372f9adfe79439d90fe74b0b5b6dea9d6/pandas/_libs/tslibs/timedeltas.pyx#L296
    abbrs = {
        'millisecond': 'millisecond',
        'ms': 'millisecond',

        'seconds': 'second',
        'second': 'second',
        'secs': 'second',
        'sec': 'second',
        's': 'second',

        'minutes': 'minute',
        'minute': 'minute',
        'mins': 'minute',
        'min': 'minute',
        'm': 'minute',

        'hours': 'hour',
        'hour': 'hour',
        'hrs': 'hour',
        'hr': 'hour',
        'h': 'hour',

        'days': 'day',
        'day': 'day',
        'd': 'day',
    }

    ms = 0
    # Finding out the leading "-"
    is_negative = False
    for i, char in enumerate(s):
        if char == '-':
            is_negative = not is_negative
        elif char == "+":
            continue
        else:
            break
    s = s[i:]
    # Continuing to find the time elements
    while s:
        pos = skip_wordbreak(s)
        s = s[pos:]

        # Example: "-  2.5  days ..."
        # Pos:         ^

        numb, pos = get_number(s)
        s = s[pos:]

        if s[0] == ":":
            # Expecting HH:MM:SS
            ms += get_hhmmss(numb + s)
            break

        # Example: "-  2.5  days ..."
        # Pos:            ^

        pos = skip_wordbreak(s)
        s = s[pos:]

        # Example: "-  2.5  days ..."
        # Pos:              ^

        abbr, pos = get_unit(s)
        s = s[pos:]

        ms += to_microseconds(**{abbr: float(numb)})

    if is_negative:
        ms = -ms
    return datetime.timedelta(microseconds=ms)

def to_microseconds(day=0, hour=0, minute=0, second=0, millisecond=0, microsecond=0) -> int:
    "Turn time components to microseconds"
    return microsecond + millisecond * 1_000 + second * int(1e+6) + minute * int(6e+7) + hour * int(3.6e+9) + day * int(8.64e+10)




import time
import datetime
from typing import Tuple

from .base import TimePeriod

def get_period_span(period:'TimePeriod') -> Tuple[datetime.datetime, datetime.datetime]:

    # To prevent circular import
    from rocketry.parse import parse_time

    if period is None:
        return TimePeriod.min, TimePeriod.max
    elif isinstance(period, str):
        period = parse_time(period)
    dt = datetime.datetime.fromtimestamp(time.time())

    interval = period.rollback(dt)
    start = interval.left
    end = interval.right
    return start, end




import datetime

from rocketry.time import (
    TimeSpanDelta
)


def test_construct():
    time = TimeSpanDelta("2:00:00", "3:00:00")
    assert time.near == datetime.timedelta(hours=2)
    assert time.far == datetime.timedelta(hours=3)

    time = TimeSpanDelta(near="2:00:00", far="3:00:00")
    assert time.near == datetime.timedelta(hours=2)
    assert time.far == datetime.timedelta(hours=3)

    time = TimeSpanDelta(near=datetime.timedelta(hours=2), far=datetime.timedelta(hours=3))
    assert time.near == datetime.timedelta(hours=2)
    assert time.far == datetime.timedelta(hours=3)

    time = TimeSpanDelta()
    assert time.near == datetime.timedelta(0)
    assert time.far is None

def test_equal():
    assert TimeSpanDelta("2 days") == TimeSpanDelta("2 days")
    assert TimeSpanDelta("2 days", "1 days") == TimeSpanDelta("2 days", "1 days")
    assert not TimeSpanDelta("2 days") == TimeSpanDelta("3 days")
    assert not TimeSpanDelta("2 days", "1 days") == TimeSpanDelta("3 days")
    assert not TimeSpanDelta("1 days", "2 days") == TimeSpanDelta("5 hours", "2 days")

def test_repr():
    assert str(TimeSpanDelta("2 days"))
    assert repr(TimeSpanDelta("2 days"))



from datetime import datetime
from rocketry.time import StaticInterval, always, never
from rocketry.pybox.time import Interval

def test_static():
    t = StaticInterval("2022-08-01", "2022-08-10")
    assert datetime.fromisoformat("2022-07-31 23:59:59") not in t
    assert datetime.fromisoformat("2022-08-01 00:00:00") in t
    assert datetime.fromisoformat("2022-08-09 12:59:00") in t
    assert datetime.fromisoformat("2022-08-10 00:00:00") not in t

    assert t.rollforward("2022-07-01 00:00:00") == Interval(
        datetime.fromisoformat("2022-08-01 00:00:00"),
        datetime.fromisoformat("2022-08-10 00:00:00"),
        closed="left"
    )
    assert t.rollback("2022-09-01 00:00:00") == Interval(
        datetime.fromisoformat("2022-08-01 00:00:00"),
        datetime.fromisoformat("2022-08-10 00:00:00"),
        closed="left"
    )
    assert t.rollforward("2022-09-01 00:00:00") == Interval(
        datetime.fromisoformat("2260-01-01 00:00:00"),
        datetime.fromisoformat("2260-01-01 00:00:00"),
        closed="left"
    )

    assert t.rollforward("2022-08-03 00:00:00") == Interval(
        datetime.fromisoformat("2022-08-03 00:00:00"),
        datetime.fromisoformat("2022-08-10 00:00:00"),
        closed="left"
    )
    assert t.rollback("2022-08-03 00:00:00") == Interval(
        datetime.fromisoformat("2022-08-01 00:00:00"),
        datetime.fromisoformat("2022-08-03 00:00:00"),
        closed="left"
    )

def test_always():
    assert datetime(2022, 1, 1) in always
    assert always.rollforward(datetime(2022, 1, 1)) == Interval(
        datetime(2022, 1, 1, 0, 0),
        datetime(2260, 1, 1, 0, 0),
        closed='left'
    )
    assert always.rollback(datetime(2022, 1, 1)) == Interval(
        datetime(1970, 1, 3, 2, 0),
        datetime(2022, 1, 1, 0, 0),
        closed='left'
    )
    assert always.is_max_interval

def test_never():
    assert datetime(2022, 1, 1) not in never
    assert never.rollforward(datetime(2022, 1, 1)) == Interval(
        datetime(2260, 1, 1, 0, 0),
        datetime(2260, 1, 1, 0, 0),
        closed='left'
    )
    assert never.rollback(datetime(2022, 1, 1)) == Interval(
        datetime(1970, 1, 3, 2, 0),
        datetime(1970, 1, 3, 2, 0),
        closed='left'
    )
    assert not never.is_max_interval



import os
import sys
import pytest
from rocketry.args import Private, SimpleArg, FuncArg, Arg, EnvArg, CliArg, Return, TerminationFlag
from rocketry.tasks import FuncTask

def test_simple():
    arg = SimpleArg("a value")
    assert arg.get_value() == "a value"

def test_func():
    def get_value():
        return 'a value'

    arg = FuncArg(get_value)
    assert arg.get_value() == "a value"

def test_arg(session):

    session.parameters['myarg'] = 'a value'
    arg = Arg('myarg')
    assert arg.get_value(session=session) == "a value"

    assert Arg('missing', default="a value").get_value(session=session) == "a value"
    with pytest.raises(KeyError):
        arg = Arg('missing')
        arg.get_value(session=session)

def test_env_arg():
    os.environ['ROCKETRY_TEST_VARIABLE'] = "a value"
    arg = EnvArg('ROCKETRY_TEST_VARIABLE')
    assert arg.get_value() == "a value"

    arg = EnvArg('ROCKETRY_NOT_SET', default="a value 2")
    assert arg.get_value() == "a value 2"

    with pytest.raises(KeyError):
        arg = EnvArg('ROCKETRY_NOT_SET')
        arg.get_value()

def test_cli_arg():
    assert CliArg.cli_args == sys.argv
    orig_arg = CliArg.cli_args
    try:
        CliArg.cli_args = ['python', 'myfile.py', '--myparam', 'a value']

        arg = CliArg('--myparam')
        assert arg.get_value() == "a value"
        assert CliArg('--invalid', default="a value").get_value() == "a value"
        with pytest.raises(KeyError):
            CliArg('--invalid').get_value()
    finally:
        CliArg.cli_args = orig_arg

def test_return(session):
    runned = FuncTask(lambda x: x, name="runned", session=session)
    unrunned = FuncTask(lambda x: x, name="unrunned", session=session)

    session.returns[runned] = 'a value'

    assert Return("runned").get_value(session=session) == "a value"
    assert Return(runned).get_value(session=session) == "a value"
    assert Return("unrunned", default="a value").get_value(session=session) == "a value"
    assert Return(unrunned, default="a value").get_value(session=session) == "a value"

    with pytest.raises(ValueError):
        Return("non_existent").get_value(session=session)

    with pytest.raises(KeyError):
        Return("unrunned").get_value(session=session)

def test_private(session):
    task = FuncTask(lambda x: x, session=session)

    value = Private("my secret")
    assert value.get_value() == "*****"
    assert value.get_value(task=task) == "my secret"
    assert str(value) == "*****"
    assert repr(value) == "Private(*****)"

@pytest.mark.parametrize("obj", [Arg('x'), SimpleArg('value'), Private('value'), FuncArg(lambda: None), Return('a_task'), TerminationFlag()])
def test_no_errors(obj, session):
    task = FuncTask(lambda: None, name="a_task", session=session)
    session.returns[task] = None
    session.parameters['x'] = None
    obj.get_value(dont_use="something", session=session, task=task)
    str(obj)
    repr(obj)




import pytest
from rocketry.pybox import query

def test_str_statement():
    qry = query.parser.from_dict({
        'mydate$min': '2021-07-01',
        'mydate$max': '2021-07-15',
        'mykey1': 10,
        'mystring$regex': r'all of .+',
        'mykey2$not': 'not this',
    })
    assert str(qry) == "((<mydate> >= '2021-07-01') & (<mydate> <= '2021-07-15') & (<mykey1> == 10) & re.match('all of .+', <mystring>) & (<mykey2> != 'not this'))"

@pytest.mark.parametrize('obj,expected',
    [
        pytest.param(query.Any(query.Key('mykey1') == 1, query.Key('mykey2') == 2), '((<mykey1> == 1) | (<mykey2> == 2))', id="any"),
        pytest.param(query.All(query.Key('mykey1') != 1, query.Key('mykey2') != 2), '((<mykey1> != 1) & (<mykey2> != 2))', id="all"),
        pytest.param(query.Key('mykey') < 1, '(<mykey> < 1)', id="less"),
        pytest.param(query.Key('mykey') <= 1, '(<mykey> <= 1)', id="less equal"),
        pytest.param(query.Key('mykey') > 1, '(<mykey> > 1)', id="greater"),
        pytest.param(query.Key('mykey') >= 1, '(<mykey> >= 1)', id="greater equal"),
        pytest.param(~(query.Key('mykey') == 1), '~(<mykey> == 1)', id="not"),
        pytest.param(query.Regex(query.Key('mykey'), r'.+ is good'), "re.match('.+ is good', <mykey>)", id="regex"),
    ]
)
def test_expr(obj, expected):
    assert str(obj) == expected

@pytest.mark.parametrize('obj,data,expected',
    [
        pytest.param(query.Any(query.Key('mykey1') == 1, query.Key('mykey2') == 2), {'mykey1': 1, 'mykey2': 2}, True, id="any-true"),
        pytest.param(query.All(query.Key('mykey1') != 1, query.Key('mykey2') != 2), {'mykey1': 2, 'mykey2': 1}, True, id="all-true"),
        pytest.param(query.Key('mykey') < 1, {'mykey': 0}, True, id="less true"),
        pytest.param(query.Key('mykey') <= 1, {'mykey': 1}, True, id="less-equal true"),
        pytest.param(query.Key('mykey') > 1, {'mykey': 2}, True, id="greater true"),
        pytest.param(query.Key('mykey') >= 1, {'mykey': 1}, True, id="greater-equal true"),
        pytest.param(~(query.Key('mykey') == 1), {'mykey': 2}, True, id="not true"),
    ]
)
def test_match(obj, data, expected):
    assert obj.match(data) if expected else not obj.match(data)




import datetime
import pytest
from rocketry.pybox import query

@pytest.mark.parametrize('qry,data,expected',
    [
        pytest.param(
            query.Key('mykey') == 10,
            [
                {'mykey': 5},
                {'mykey': 10},
                {'mykey': 11},
            ],
            [{'mykey': 10}],
            id="equal"
        ),
        pytest.param(
            query.Key('mydate') == '2021-01-01',
            [
                {'mydate': datetime.datetime.fromisoformat('2020-12-31')},
                {'mydate': datetime.datetime.fromisoformat('2021-01-01')},
                {'mydate': datetime.datetime.fromisoformat('2020-01-02')},
            ],
            [{'mydate': datetime.datetime.fromisoformat('2021-01-01')}],
            id="string to datetime"
        ),
    ]
)
def test_filter(qry, data, expected):
    assert list(qry.filter(data)) == expected




import asyncio
import time
import pytest

from rocketry.args import Session
import rocketry
from rocketry.args.builtin import TerminationFlag
from rocketry.conditions.scheduler import SchedulerStarted
from rocketry.conditions import TaskStarted
from rocketry.core.time.base import TimeDelta
from rocketry.tasks import FuncTask
from rocketry.conds import true
from rocketry.exc import TaskTerminationException


@pytest.mark.parametrize("execution", ["main", "thread"])
def test_shutdown(execution, session):
    timeline = []

    def on_startup():
        timeline.append("startup")

    def call_shutdown(session:rocketry.Session=Session()):
        session.shut_down()
        timeline.append("shutdown-called")

    def on_shutdown():
        timeline.append("shutdown")


    FuncTask(on_startup, name="startup", on_startup=True, start_cond=true, execution=execution, session=session)
    FuncTask(call_shutdown, name="call-shutdown", start_cond=true, execution=execution, session=session)
    FuncTask(on_shutdown, name="shutdown", on_shutdown=True, start_cond=true, execution=execution, session=session)


    session.config.shut_cond = TaskStarted(task="call-shutdown") >= 3

    session.start()
    assert timeline == ['startup', 'shutdown-called', 'shutdown']

@pytest.mark.parametrize("doublecall", [False, True])
@pytest.mark.parametrize("execution", ["async", "thread"])
def test_shutdown_force(execution, session, doublecall):
    timeline = []

    def on_startup():
        timeline.append("startup")

    async def slow_task_async():
        await asyncio.sleep(2)

    def slow_task_thread(flag=TerminationFlag()):
        while not flag.is_set():
            time.sleep(0.001)
        if flag.is_set():
            raise TaskTerminationException()

    def call_shutdown(session:rocketry.Session=Session()):
        if doublecall:
            session.shut_down()
            session.shut_down()
        else:
            session.shut_down(force=True)
        timeline.append("shutdown-called")

    slow_func = {'async': slow_task_async, 'thread': slow_task_thread}[execution]

    FuncTask(on_startup, name="startup", on_startup=True, start_cond=true, execution=execution, session=session)
    task = FuncTask(slow_func, name="slow-task", start_cond=true, execution=execution, session=session)
    FuncTask(call_shutdown, name="call-shutdown", start_cond=true, execution=execution, session=session)
    #FuncTask(on_shutdown, name="shutdown", on_shutdown=True, start_cond=true, execution=execution, session=session)


    session.config.shut_cond = TaskStarted(task="call-shutdown") >= 3

    session.start()
    assert timeline == ['startup', 'shutdown-called']
    assert 1 == task.logger.filter_by(action="terminate").count()

@pytest.mark.parametrize("execution", ["main", "thread"])
def test_restart(execution, session):
    timeline = []
    session.config.restarting = 'recall'

    def on_startup():
        timeline.append("startup")

    def call_restart(session:rocketry.Session=Session()):
        session.restart()
        timeline.append("restart-called")

    def on_shutdown():
        timeline.append("shutdown")

    FuncTask(on_startup, name="startup", on_startup=True, execution="main", session=session)
    FuncTask(call_restart, name="restart", start_cond=true, execution=execution, session=session)
    FuncTask(on_shutdown, name="shutdown", on_shutdown=True, execution="main", session=session)

    session.config.shut_cond = (TaskStarted(task="shutdown") >= 2) | ~SchedulerStarted(period=TimeDelta("60 second"))

    session.start()
    assert timeline == ['startup', 'restart-called', 'shutdown', 'startup', 'restart-called', 'shutdown', 'startup', 'shutdown']



from pathlib import Path

def create_file(file, content):
    path = Path(file)
    path.parent.mkdir(exist_ok=True)
    #with open(path, "w") as f:
    #    f.write(content)
    path.write_text(content)

def delete_file(file):
    Path(file).unlink()



import calendar
import datetime
import re
from dataclasses import dataclass
from typing import ClassVar, List

import dateutil

from rocketry.core.time.anchor import AnchoredInterval
from rocketry.core.time.base import TimeInterval
from rocketry.pybox.time import datetime_to_dict, to_microseconds
from rocketry.pybox.time.interval import Interval

@dataclass(frozen=True, init=False)
class TimeOfSecond(AnchoredInterval):
    """Time interval anchored to second cycle of a clock

    min: 0 microsecond
    max: 999999 microsecond

    """

    _scope: ClassVar[str] = "second"

    _scope_max: ClassVar[int] = to_microseconds(second=1)
    _unit_resolution: ClassVar[int] = to_microseconds(millisecond=1)
    _unit_names: ClassVar[List[str]] = [str(i) for i in range(1000)] # 00, 01 etc. till 59

    def anchor_float(self, i, **kwargs):
        return to_microseconds(millisecond=i)

    def anchor_int(self, i, **kwargs):
        if not 0 <= i <= 1000:
            raise ValueError(f"Invalid value: {i}. Allowed: 0-1000")
        return super().anchor_int(i, **kwargs)

    def anchor_str(self, s, **kwargs):
        # ie. 30.123
        res = float(s)
        return to_microseconds(millisecond=res)

@dataclass(frozen=True, init=False)
class TimeOfMinute(AnchoredInterval):
    """Time interval anchored to minute cycle of a clock

    min: 0 seconds, 0 microsecond
    max: 59 seconds, 999999 microsecond
    """

    _scope: ClassVar[str] = "minute"

    _scope_max: ClassVar[int] = to_microseconds(minute=1)
    _unit_resolution: ClassVar[int] = to_microseconds(second=1)
    _unit_names: ClassVar[List[str]] = [f"{i:02d}" for i in range(60)] # 00, 01 etc. till 59

    def anchor_float(self, i, **kwargs):
        return to_microseconds(second=i)

    def anchor_int(self, i, **kwargs):
        if not 0 <= i <= 59:
            raise ValueError(f"Invalid value: {i}. Allowed: 0-59")
        return super().anchor_int(i, **kwargs)

    def anchor_str(self, s, **kwargs):
        # ie. 30.123
        res = re.search(r"(?P<second>[0-9][0-9])([.](?P<microsecond>[0-9]{0,6}))?", s, flags=re.IGNORECASE)
        if res:
            res = res.groupdict()
            if res["microsecond"] is not None:
                res["microsecond"] = res["microsecond"].ljust(6, "0")
            return to_microseconds(**{key: int(val) for key, val in res.items() if val is not None})


@dataclass(frozen=True, init=False)
class TimeOfHour(AnchoredInterval):
    """Time interval anchored to hour cycle of a clock

    min: 0 minutes, 0 seconds, 0 microsecond
    max: 59 minutes, 59 seconds, 999999 microsecond

    Example:
        # From 15 past to half past
        TimeOfHour("15:00", "30:00")
    """
    _scope: ClassVar[str] = "hour"
    _scope_max: ClassVar[int] = to_microseconds(hour=1)
    _unit_resolution: ClassVar[int] = to_microseconds(minute=1)
    _unit_names: ClassVar[List[str]] = [f"{i:02d}:00" for i in range(60)] # 00:00, 01:00 etc. till 59:00

    def anchor_int(self, i, **kwargs):
        if not 0 <= i <= 59:
            raise ValueError(f"Invalid minute: {i}. Minute is from 0 to 59")
        return super().anchor_int(i, **kwargs)

    def anchor_str(self, s, **kwargs):
        # ie. 12:30.123
        res = re.search(r"(?P<minute>[0-9][0-9]):(?P<second>[0-9][0-9])([.](?P<microsecond>[0-9]{0,6}))?", s, flags=re.IGNORECASE)
        if res:
            res = res.groupdict()
            if res["microsecond"] is not None:
                res["microsecond"] = res["microsecond"].ljust(6, "0")
            return to_microseconds(**{key: int(val) for key, val in res.items() if val is not None})

        res = re.search(r"(?P<n>[0-4]) ?(quarter|q)", s, flags=re.IGNORECASE)
        if res:
            # ie. "1 quarter"
            n_quarters = int(res["n"])
            return (self._scope_max + 1) / 4 * n_quarters - 1
        raise ValueError(f"Invalid value: {repr(s)}")

@dataclass(frozen=True, init=False)
class TimeOfDay(AnchoredInterval):
    """Time interval anchored to day cycle of a clock

    min: 0 hours, 0 minutes, 0 seconds, 0 microsecond
    max: 23 hours, 59 minutes, 59 seconds, 999999 microsecond

    Example:
        # From 10 o'clock to 15 o'clock
        TimeOfDay("10:00", "15:00")
    """
    _scope: ClassVar[str] = "day"
    _scope_max: ClassVar[int] = to_microseconds(day=1)
    _unit_resolution: ClassVar[int] = to_microseconds(hour=1)
    _unit_names: ClassVar[List[str]] = [f"{i:02d}:00" for i in range(24)] # 00:00, 01:00, 02:00 etc. till 23:00

    def anchor_int(self, i, **kwargs):
        if not 0 <= i <= 23:
            raise ValueError(f"Invalid hour: {i}. Day is from 00 to 23")
        return super().anchor_int(i, **kwargs)

    def anchor_str(self, s, **kwargs):
        # ie. "10:00:15"
        dt = dateutil.parser.parse(s)
        d = datetime_to_dict(dt)
        components = ("hour", "minute", "second", "microsecond")
        return to_microseconds(**{key: int(val) for key, val in d.items() if key in components})

    def anchor_dt(self, dt, **kwargs):
        "Turn datetime to microseconds according to the scope (by removing higher time elements)"
        d = datetime_to_dict(dt)
        d = {
            key: val
            for key, val in d.items()
            if key in ("hour", "minute", "second", "microsecond")
        }
        return to_microseconds(**d)

@dataclass(frozen=True, init=False)
class TimeOfWeek(AnchoredInterval):
    """Time interval anchored to week cycle

    min: Monday, 0 hours, 0 minutes, 0 seconds, 0 microsecond
    max: Sunday, 23 hours, 59 minutes, 59 seconds, 999999 microsecond

    Example:
        # From Monday 3 PM to Wednesday 4 PM
        TimeOfWeek("Mon 15:00", "Wed 16:00")
    """
    _scope: ClassVar[str] = "week"
    _scope_max: ClassVar[int] = to_microseconds(day=7) # Sun day end of day
    _unit_resolution: ClassVar[int] = to_microseconds(day=1)
    _unit_names: ClassVar[List[str]] = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']

    _unit_mapping = {
        **dict(zip(range(1, 8), range(7))),

        # English
        **{
            day: i
            for i, day in enumerate(['mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun'])
        },
        **{
            day: i
            for i, day in enumerate(['monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday'])
        },
    }

    def anchor_int(self, i, **kwargs):
        # The axis is from 0 to 6 times microseconds per day
        # but if start/end is passed as int, it's considered from 1-7
        # (Monday is 1)
        if not 1 <= i <= 7:
            raise ValueError("Invalid day of week. Week day is from 1 (Monday) to 7 (Sunday).")
        return super().anchor_int(i-1, **kwargs)

    def anchor_str(self, s, side=None, **kwargs):
        # Allowed:
        #   "Mon", "Monday", "Mon 10:00:00"
        res = re.search(r"(?P<dayofweek>[a-z]+) ?(?P<time>.*)", s, flags=re.IGNORECASE)
        comps = res.groupdict()
        dayofweek = comps.pop("dayofweek")
        time = comps.pop("time")
        try:
            nth_day = self._unit_mapping[dayofweek.lower()]
        except KeyError:
            raise ValueError(f"Invalid day of week: {dayofweek}")

        # TODO: TimeOfDay.anchor_str as function
        if not time:
            microseconds = to_microseconds(day=1) if side == "end" else 0
        else:
            microseconds = TimeOfDay().anchor_str(time)

        return to_microseconds(day=1) * nth_day + microseconds

    def anchor_dt(self, dt, **kwargs):
        "Turn datetime to microseconds according to the scope (by removing higher time elements)"
        d = datetime_to_dict(dt)
        d = {
            key: val
            for key, val in d.items()
            if key in ("hour", "minute", "second", "microsecond")
        }
        dayofweek = dt.weekday()
        return to_microseconds(**d) + dayofweek * to_microseconds(day=1)


@dataclass(frozen=True, init=False)
class TimeOfMonth(AnchoredInterval):
    """Time interval anchored to day cycle of a clock

    min: first day of month, 0 hours, 0 minutes, 0 seconds, 0 microsecond
    max: last day of month, 23 hours, 59 minutes, 59 seconds, 999999 microsecond

    Example:
        # From 10 o'clock to 15 o'clock
        TimeOfMonth("1st", "5th")
    """
    # TODO: Support for month end (ie. last 5th day of month to last 2nd)
    # Could be implemented by allowing minus _start and minus _end
    #   rollforward/rollback/contains would need slight changes

    _scope: ClassVar[str] = "year"
    _scope_max: ClassVar[int] = to_microseconds(day=31) # 31st end of day
    _unit_resolution: ClassVar[int] = to_microseconds(day=1)
    _unit_names: ClassVar[List[str]] = ["1st", "2nd", "3rd"] + [f"{i}th" for i in range(4, 32)]
     # NOTE: Floating
    # TODO: ceil end and implement reversion (last 5th day)

    def anchor_int(self, i, **kwargs):
        if not 1 <= i <= 31:
            raise ValueError("Invalid day of month. Day of month can be from 1 to 31")
        # We allow passing days from 1-31 but the axis is built on starting from zero
        i -= 1
        return super().anchor_int(i, **kwargs)

    def anchor_str(self, s, side=None, **kwargs):
        # Allowed:
        #   "5th", "1st", "5", "3rd 12:30:00"
        #   TODO: "first 5th", "last 5th", "last 3rd 10:00:00"
        res = re.search(r"(?P<dayofmonth>[0-9]+)(st|nd|rd|th|[.])? ?(?P<time>.*)", s, flags=re.IGNORECASE)
        comps = res.groupdict()
        dayofmonth = comps.pop("dayofmonth")
        time = comps.pop("time")
        nth_day = int(dayofmonth)

        # TODO: TimeOfDay.anchor_str as function
        ceil_time = not time and side == "end"
        if ceil_time:
            # If time is not defined and the end
            # is being anchored, the time is ceiled.

            # If one says 'thing X was organized between
            # 15th and 17th of July', the sentence
            # includes 17th till midnight.
            microseconds = to_microseconds(day=1)
        elif time:
            microseconds = TimeOfDay().anchor_str(time)
        else:
            microseconds = 0

        return to_microseconds(day=1) * (nth_day - 1) + microseconds

    def anchor_dt(self, dt, **kwargs):
        "Turn datetime to microseconds according to the scope (by removing higher time elements)"
        d = datetime_to_dict(dt)
        d = {
            key: val
            for key, val in d.items()
            if key in ("day", "hour", "minute", "second", "microsecond")
        }
        if "day" in d:
            # Day (of month) does not start from 0 (but from 1)
            d["day"] = d["day"] - 1

        return to_microseconds(**d)

    def get_scope_forward(self, dt):
        n_days = calendar.monthrange(dt.year, dt.month)[1]
        return to_microseconds(day=1) * n_days

    def get_scope_back(self, dt):
        month = 12 if dt.month == 1 else dt.month - 1
        year = dt.year - 1 if dt.month == 1 else dt.year
        n_days = calendar.monthrange(year, month)[1]
        return to_microseconds(day=1) * n_days

@dataclass(frozen=True, init=False)
class TimeOfYear(AnchoredInterval):
    """Time interval anchored to day cycle of a clock

    min: 1st Jan, 0 hours, 0 minutes, 0 seconds, 0 microsecond
    max: 31st Dec, 23 hours, 59 minutes, 59 seconds, 999999 microsecond

    Example:
        # From 10 o'clock to 15 o'clock
        TimeOfYear("Jan", "Feb")
    """

    # We take the longest year there is and translate all years to that
    # using first the month and then the day of month

    _scope: ClassVar[str] = "year"
    _scope_max: ClassVar[int] = to_microseconds(day=1) * 366
    _unit_names: ClassVar[List[str]] = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']

    _unit_mapping: ClassVar = {
        **dict(zip(range(12), range(12))),

        # English
        **{day.lower(): i for i, day in enumerate(['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December'])},
        **{day.lower(): i for i, day in enumerate(_unit_names)},
    }

    _month_start_mapping: ClassVar = {
        0: 0, # January
        1: to_microseconds(day=31), # February (31 days from year start)
        2: to_microseconds(day=60), # March (31 + 29, leap year has 29 days in February)
        3: to_microseconds(day=91), # April (31 + 29 + 31)
        4: to_microseconds(day=121), # May (31 + 29 + 31 + 30)
        5: to_microseconds(day=152), # June
        6: to_microseconds(day=182), # July
        7: to_microseconds(day=213), # August

        8: to_microseconds(day=244), # September
        9: to_microseconds(day=274), # October
        10: to_microseconds(day=305), # November
        11: to_microseconds(day=335), # December
        12: to_microseconds(day=366), # End of the year (on leap years)
    }
    # Reverse the _month_start_mapping to microseconds to month num
    _year_start_mapping: ClassVar = dict((v, k) for k, v in _month_start_mapping.items())

    # NOTE: Floating

    def anchor_str(self, s, side=None, **kwargs):
        # Allowed:
        #   "January", "Dec", "12", "Dec last 5th 10:00:00"
        res = re.search(r"(?P<monthofyear>[a-z]+) ?(?P<day_of_month>.*)", s, flags=re.IGNORECASE)
        comps = res.groupdict()
        monthofyear = comps.pop("monthofyear") # This is jan, january
        day_of_month_str = comps.pop("day_of_month")
        nth_month = self._unit_mapping[monthofyear.lower()]

        ceil_time = not day_of_month_str and side == "end"
        if ceil_time:
            # If time is not defined and the end
            # is being anchored, the time is ceiled.

            # If one says 'thing X was organized between
            # May and June', the sentence includes
            # time between 1st of May to 30th of June.
            return self._month_start_mapping[nth_month+1] - 1
        elif day_of_month_str:
            microseconds = TimeOfMonth().anchor_str(day_of_month_str)
        else:
            microseconds = 0

        return self._month_start_mapping[nth_month] + microseconds

    def to_timepoint(self, ns:int):
        "Turn microseconds to the period's timepoint"
        # Ie. Monday --> Monday 00:00 to Monday 24:00
        # By default assumes linear scale (like week)
        # but can be overridden for non linear such as year
        month_num = self._year_start_mapping[ns]
        return self._month_start_mapping[month_num + 1] - 1

    def anchor_int(self, i, side=None, **kwargs):
        # i is the month (Jan = 1)
        # The axis is from 0 to 365 * microseconds per day
        if not 1 <= i <= 12:
            raise ValueError(f"Invalid month: {i} (Jan is 1 and Dec is 12)")
        i -= 1
        if side == "end":
            return self._month_start_mapping[i+1] - 1
        return self._month_start_mapping[i]

    def anchor_dt(self, dt, **kwargs):
        "Turn datetime to microseconds according to the scope (by removing higher time elements)"
        dt_dict = datetime_to_dict(dt)
        d = {
            key: val
            for key, val in dt_dict.items()
            if key in ("day", "hour", "minute", "second", "microsecond")
        }
        nth_month = dt_dict["month"] - 1
        if "day" in d:
            # Day (of month) does not start from 0 (but from 1)
            d["day"] = d["day"] - 1
        return self._month_start_mapping[nth_month] + to_microseconds(**d)


@dataclass(frozen=True, init=False)
class RelativeDay(TimeInterval):
    """Specific day

    Examples:
    ---------
        Day("today")
        Day("yesterday")
        Day("yesterday")
    """

    offsets: ClassVar = {
        "today": datetime.timedelta(),
        "yesterday": datetime.timedelta(days=1),
        "the_day_before":datetime.timedelta(days=2),
        #"first_day_of_year": get_first_day_of_year,
    }

    min_time: ClassVar[datetime.time] = datetime.time.min
    max_time: ClassVar[datetime.time] = datetime.time.max

    def __init__(self, day, *, start_time=None, end_time=None):
        object.__setattr__(self, "day", day)
        object.__setattr__(self, "start_time", self.min_time if start_time is None else start_time)
        object.__setattr__(self, "end_time", self.max_time if end_time is None else end_time)

    def rollback(self, dt):
        offset = self.offsets[self.day]
        dt = dt - offset
        return Interval(
            datetime.datetime.combine(dt, self.start_time),
            datetime.datetime.combine(dt, self.end_time)
        )

    def rollforward(self, dt):
        raise AttributeError("RelativeDay has no next day")

    def __repr__(self):
        args_str = str(self.day)
        if self.start_time != self.min.date():
            args_str += f', start_time={self.start_time}'
        if self.end_time != self.max.date():
            args_str += f', end_time={self.end_time}'

        return f"RelativeDay({args_str})"



import datetime
import logging
import os
import sys
import time
from pathlib import Path

import pytest
from dateutil.parser import parse as parse_datetime

from redbird.logging import RepoHandler
from redbird.repos import MemoryRepo

import rocketry
from rocketry import Session
from rocketry.core.hook import clear_hooks
from rocketry.log.log_record import MinimalRecord

# add helpers to path
sys.path.append(os.path.join(os.path.dirname(__file__), 'helpers'))

# Utils
def get_node_id(request):
    components = request.node.nodeid.split("::")
    filename = components[0]
    test_class = components[1] if len(components) == 3 else None
    test_func_with_params = components[-1]
    test_func = test_func_with_params.split('[')[0]

    filename = filename.replace(".py", "").replace("/", "-")
    if test_class:
        return f'{filename}-{test_class}-{test_func_with_params}'
    return f'{filename}-{test_func_with_params}'

def pytest_sessionstart(session):
    """
    Called after the Session object has been created and
    before performing collection and entering the run test loop.
    """
    rocketry.session.config.debug = True


def copy_file_to_tmpdir(tmpdir, source_file, target_path):
    target_path = Path(target_path)
    source_path = Path(os.path.dirname(__file__)) / "test_files" / source_file

    fh = tmpdir.join(target_path.name)
    with open(source_path) as f:
        fh.write(f.read())
    return fh

@pytest.fixture(autouse=True)
def sys_paths():
    orig_sys_paths = sys.path.copy()
    yield
    # Setting back the original sys paths so they are not
    # carried over different tasks.
    sys.path = orig_sys_paths

@pytest.fixture
def script_files(tmpdir):
    for folder in Path("scripts").parts:
        tmpdir = tmpdir.mkdir(folder)

    copy_file_to_tmpdir(tmpdir, source_file="succeeding_script.py", target_path="scripts/succeeding_script.py")
    copy_file_to_tmpdir(tmpdir, source_file="failing_script.py", target_path="scripts/failing_script.py")
    copy_file_to_tmpdir(tmpdir, source_file="parameterized_script.py", target_path="scripts/parameterized_script.py")
    copy_file_to_tmpdir(tmpdir, source_file="parameterized_kwargs_script.py", target_path="scripts/parameterized_kwargs_script.py")
    copy_file_to_tmpdir(tmpdir, source_file="syntax_error_script.py", target_path="scripts/syntax_error_script.py")


@pytest.fixture(scope="function", autouse=True)
def session():
    session = Session(config={
        "debug": True,
        "silence_task_prerun": False,
        "silence_task_logging": False,
        "silence_cond_check": False,
        "cycle_sleep": 0.001,
        "task_execution": "process",
    }, delete_existing_loggers=True)
    rocketry.session = session
    session.set_as_default()

    task_logger = logging.getLogger(session.config.task_logger_basename)
    task_logger.handlers = [
        RepoHandler(repo=MemoryRepo(model=MinimalRecord)),
        #logging.StreamHandler(sys.stdout)
    ]
    task_logger.setLevel(logging.INFO)

    # enable logger
    # Some tests may disable especially scheduler logger if logging config has
    # "disable_existing_loggers" as True and missing scheduler logger
    logging.getLogger(session.config.task_logger_basename).disabled = False
    logging.getLogger(session.config.scheduler_logger_basename).disabled = False

    # Clear hooks
    clear_hooks()
    return session

@pytest.fixture(scope="session", autouse=True)
def set_loggers():
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(action)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)

    task_logger = logging.getLogger(rocketry.session.config.task_logger_basename)
    task_logger.addHandler(handler)

    yield session

class mockdatetime(datetime.datetime):
    _freezed_datetime = None
    @classmethod
    def now(cls):
        return cls._freezed_datetime

@pytest.fixture
def mock_datetime_now(monkeypatch):
    """Monkey patch datetime.datetime.now
    Returns a function that takes datetime as string as input
    and sets that to datetime.datetime.now()"""
    import datetime
    class mockdatetime(datetime.datetime):
        _freezed_datetime = None
        @classmethod
        def now(cls):
            return cls._freezed_datetime

    def wrapper(dt):
        dt = parse_datetime(dt)

        # Mock datetime.datetime
        mockdatetime._freezed_datetime = mockdatetime(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.microsecond)
        monkeypatch.setattr(datetime, 'datetime', mockdatetime)

        # Mock time.time
        mocktime = dt.timestamp()
        monkeypatch.setattr(time, 'time', lambda: mocktime)

    yield wrapper

@pytest.fixture
def mock_time(monkeypatch):
    """Monkey patch time.time
    Returns a function that takes datetime as string as input
    and sets that to time.time()"""

    def wrapper(dt):
        mocktime = parse_datetime(dt).timestamp()
        monkeypatch.setattr(time, 'time', lambda: mocktime)

    return wrapper

@pytest.fixture
def mock_pydatetime(mock_time, mock_datetime_now):
    """Monkey patch time.time & datetime.datetime.now
    Returns a function that takes datetime as string as input
    and sets that to time.time() and datetime.datetime.now()"""

    def wrapper(dt):
        mock_time(dt)
        mock_datetime_now(dt)
    return wrapper

# Mongo Database
@pytest.fixture(scope="function")
def mongo_conn_str():
    conf_path = Path("rocketry/test/private.yaml")
    pytest.importorskip("pymongo")
    if not conf_path.is_file():
        pytest.skip("Missing Mongo connection")
    import yaml

    with open(conf_path, 'r') as f:
        conf = yaml.safe_load(f)
    return conf["mongodb"]["conn_str"]

@pytest.fixture(scope="function")
def mongo_client(mongo_conn_str):
    import pymongo
    return pymongo.MongoClient(mongo_conn_str)

@pytest.fixture(scope="function")
def collection(request, mongo_client):

    db_name = "pytest"
    col_name = get_node_id(request)

    collection = mongo_client[db_name][col_name]

    # Empty the collection
    collection.delete_many({})

    yield collection



from .base import (
    #Query,
    Key,
    All, Any, Not,
    Equal, NotEqual, Greater, GreaterEqual, Less, LessEqual
)
from .parse import Parser
from .string import Regex
parser = Parser()
from_dict_string = parser.from_dict




from . import task

from . import time
from . import schedule
from . import condition

from .parameters import Parameters, BaseArgument
from .schedule import Scheduler
from .task import Task


from .condition import BaseCondition
from .time import TimePeriod




import re

import pytest

from rocketry.core import BaseCondition
from rocketry.conditions import FuncCond
from rocketry.parse.condition import parse_condition

def test_func_cond():

    @FuncCond(syntax="is foo")
    def is_foo():
        return True

    cond = parse_condition("is foo")
    assert isinstance(cond, BaseCondition)
    assert bool(cond)

def test_func_cond_with_kwargs():

    @FuncCond(syntax=re.compile("is foo (?P<myval>true|false)"))
    def is_foo(myval):
        return True if myval == "true" else False if myval == "false" else None

    cond_true = parse_condition("is foo true")
    assert isinstance(cond_true, BaseCondition)
    assert bool(cond_true)

    cond_false = parse_condition("is foo false")
    assert isinstance(cond_false, BaseCondition)
    assert not bool(cond_false)

    assert cond_true is not cond_false

def test_func_cond_pass():

    @FuncCond()
    def is_foo(myval):
        return True if myval == "true" else False if myval == "false" else None

    # Test cond API
    cond = is_foo(myval="true")
    assert cond.kwargs == {"myval": "true"}

    cond = is_foo("true")
    assert cond.args == ("true",)

    assert is_foo(myval="true").observe()
    assert is_foo("true").observe()
    assert not is_foo(myval="false").observe()
    assert not is_foo("false").observe()

def test_func_cond_fail():
    cond = FuncCond()
    with pytest.raises(ValueError):
        # Has no func specified so should fail
        cond(value="asd")




import pytest
from rocketry.core.time.anchor import AnchoredInterval

# Test no unexpected errors in all
@pytest.mark.parametrize("method", ["__str__", "__repr__"])
@pytest.mark.parametrize("cls", AnchoredInterval.__subclasses__())
def test_magic_noerror(method, cls):
    obj = cls()
    getattr(obj, method)()



import re

import pytest

from rocketry.conditions import TaskCond
from rocketry.conditions import SchedulerCycles, SchedulerStarted, TaskStarted
from rocketry import Session
from rocketry.tasks import FuncTask

N_PARSERS = len(Session._cls_cond_parsers)

def is_foo(status):
    print(f"evaluating: {status}")
    if status == "true":
        return True
    else:
        return False

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_taskcond_true(session, execution):
    assert session._cond_cache == {}

    cond = TaskCond(syntax=re.compile(r"is foo (?P<status>.+)"), start_cond="every 1 min", active_time="past 10 seconds", execution=execution, session=session)
    cond(is_foo)

    task = FuncTask(lambda: None, start_cond="is foo true", name="a task", execution="main", session=session)

    # Test that there is only one more cond parser
    assert len(session._cond_parsers) == N_PARSERS + 1

    session.config.shut_cond = (TaskStarted(task="a task") >= 2) | ~SchedulerStarted(period="past 5 seconds")
    session.start()

    records = list(map(lambda e: e.dict(exclude={'created'}), session.get_task_log()))
    history_task = [
        rec
        for rec in records
        if rec['task_name'] == "a task"
    ]
    assert history_task == [
        {"task_name": "a task", "action": "run"},
        {"task_name": "a task", "action": "success"},
        {"task_name": "a task", "action": "run"},
        {"task_name": "a task", "action": "success"},
    ]

    # Check cond task
    cond_tasks = [task for task in session.tasks if task.name.startswith("_condition")]
    assert len(cond_tasks) == 1

    cond_task = cond_tasks[0]
    history_check = [
        rec
        for rec in records
        if rec['task_name'] == cond_task.name
    ]
    assert history_check == [
        {"task_name": cond_task.name, "action": "run"},
        {"task_name": cond_task.name, "action": "success"},
    ]

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_taskcond_false(session, execution):
    assert session._cond_cache == {}

    cond = TaskCond(syntax=re.compile(r"is foo (?P<status>.+)"), start_cond="every 1 min", active_time="past 10 seconds", execution=execution, session=session)
    cond(is_foo)

    # Test that there is only one more cond parser
    assert len(session._cond_parsers) == N_PARSERS + 1

    task = FuncTask(lambda: None, start_cond="is foo false", name="a task", execution="main", session=session)

    session.config.shut_cond = SchedulerCycles() >= 3
    session.start()

    records = list(map(lambda e: e.dict(exclude={'created'}), session.get_task_log()))
    history_task = [
        rec
        for rec in records
        if rec["task_name"] == "a task"
    ]
    assert history_task == []

    # Check cond task
    cond_tasks = [task for task in session.tasks if task.name.startswith("_condition")]
    assert len(cond_tasks) == 1

    cond_task = cond_tasks[0]
    history_check = [
        rec
        for rec in records
        if rec["task_name"] == cond_task.name
    ]
    assert history_check == [
        {"task_name": cond_task.name, "action": "run"},
        {"task_name": cond_task.name, "action": "success"},
    ]



from rocketry.args.builtin import Session
from rocketry.core.condition.base import BaseComparable, BaseCondition
from rocketry.core.time.utils import get_period_span


class SchedulerCycles(BaseComparable):
    """Condition for whether the scheduler have had
    more/less/equal given amount of cycles of executing
    tasks.
    """

    def get_measurement(self, session=Session()) -> int:
        n_cycles = session.scheduler.n_cycles
        return n_cycles

    def __str__(self):
        if hasattr(self, "_str"):
            return self._str
        comps = {
            "__eq__": "has",
            "__gt__": "has more than",
            "__lt__": "has less than",
            "__ge__": "has more or equal than",
            "__le__": "has less or equal than",
        }
        s = "scheduler"
        for key, val in self._comps.items():
            s = s + " " + comps.get(key, key) + " " + str(val) + " cycles"
        return s

    @classmethod
    def from_magic(cls, **kwargs):
        kwargs = {key: int(value) for key, value in kwargs.items()}
        return super(SchedulerCycles, cls).from_magic(**kwargs)

class SchedulerStarted(BaseCondition):
    """Condition for whether the scheduler had started
    in given period.

    Examples
    --------

    **Parsing example:**

    >>> from rocketry.parse import parse_condition
    >>> parse_condition("scheduler started 10 minutes ago")
    SchedulerStarted(period=TimeDelta('10 minutes'))

    >>> parse_condition("scheduler has run over 10 minutes")
    ~SchedulerStarted(period=TimeDelta('10 minutes'))
    """

    def __init__(self, period=None):
        self.period = period
        super().__init__()

    def get_state(self, session=Session()) -> bool:
        start, end = get_period_span(self.period)
        dt = session.scheduler.startup_time
        return start <= dt <= end

    def __str__(self):
        if hasattr(self, "_str"):
            return self._str
        return f"scheduler {self.period}"



from collections.abc import Iterable

class Visitor:

    """A visitor class to visit elements of nested containers

    Example:
    --------
        v = Visitor(
            visit_types=(list, tuple)
        )
        # Apply a function to each element (return None)
        v.apply([[11, 12, 13], 2, [311, [321, 322, 323], 33], 3], print)

        # Flatten the container
        v.flatten([[11, 12, 13], 2, [311, [321, 322, 323], 33], 3])
        >>> [11, 12, 13, 2, 311, 321, 322, 323, 33, 3]
    """

    def __init__(self, use_attr=None, use_func=None,
                 visit_types=None, visit_func=None, visit_with_attr=None, visit_iterables=None,
                 iter_attr=None, iter_item=None, iter_func=None):

        # Iteration:
        #   iter([1, 2, 3])
        #   myobj.things
        #   my_generator([1,2,3])
        #   myobj["things"]

        # Iteration check
        #
        self.use_attr = use_attr
        self.use_func = use_func
        self.iter_func = iter_func

        self.visit_funcs = []
        if visit_types is not None:
            self.visit_funcs.append(lambda cont: isinstance(cont, visit_types))

    def flatten(self, cont):
        """Turn nested container to flat container

        Example:
        --------
            v = Visitor(visit_types=(list,))
            v.flatten([[1,2,3], 4, [5,6,7]])
            >>> [1,2,3,4,5,6,7]
        """
        if not self.is_visitable(cont):
            # End up here only if first level of recursion
            return [cont]

        tgtcont = []
        for key, subcont in self.iter(cont):

            if self.is_visitable(subcont):
                out = self.flatten(subcont)
            else:
                out = [subcont]
            tgtcont += out
        return tgtcont

    def assign_elements(self, cont, func):
        "Assign function to the elements of the container"
        # cont: [1, [21, 22, 23], 3]
        #   iter 0:
        #       cont[0] = func(1)
        #   iter 1:
        #       subcont = [21, 22, 23]
        #       ....
        # cont: {"a": 1, "b": {"ba": 1, "bb": 2, "bc": 3}, "c": 2}
        #   iter 0:
        #       cont["a"] = func(1)
        #   iter 1:
        #       subcont = {"ba": 1, "bb": 2, "bc": 3}
        #       ....
        if not self.is_visitable(cont):
            # End up here only if first level of recursion
            raise TypeError(f"Values of {cont} cannot be set anywhere. The object must be a visitable")

        for key, obj in self.iter(cont):
            if self.is_visitable(obj):
                self.assign_elements(obj, func)
            else:
                value = func(obj)
                self._assign_value(cont, key, value)

    def _assign_value(self, cont, key, value):
        # list: cont=[1, 2, 3], key=0, value=5
        #   cont[0] = 5
        # dict: cont=[1, 2, 3], key=0, value=5

        if self.use_attr:
            # cont = Namespace(myattr=[1,2,3]) (key=0, value=-1)
            getattr(cont, self.use_attr)[key] = value
        else:
            # cont = [1,2,3] (key=0, value=-1)
            # cont = {"a":1, "b":2} (key="a", value=-1)
            cont[key] = value

    def reduce(self, cont, func):
        args = []
        for key, subobj in self.iter(cont):
            if not self.is_visitable(subobj):
                # subobj = 5
                args.append(subobj)
            elif self.has_sub_containers(subobj):
                # subobj = [1, [2, 3], 4]
                value = self.reduce(subobj, func)
                args.append(value)
            else:
                # # subobj = [1, 2, 3, 4]
                subargs = [val for key, val in self.iter(subobj)]
                value = func(*subargs)
                args.append(value)
        # Now should be inner
        # cont = [3, 3]
        return func(*args)

    def has_sub_containers(self, obj):
        if not self.is_visitable(obj):
            return False
        return any(
            self.is_visitable(subobj)
            for key, subobj in self.iter(obj)
        )

    def apply(self, cont, func):
        "Apply a function to each container"
        func(cont)
        for key, subcont in self.iter(cont):
            if self.is_visitable(subcont):
                # Container
                self.apply(subcont, func)
            else:
                # Actual element
                pass

    def iter(self, obj):
        # List:
        #   [1,2,3]
        #   1, 2, 3
        # Dict:
        #   {"a": 1, "b": 2, "c": 3}
        #   1, 2, 3

        cont = self._get_container(obj)

        if self.iter_func:
            return self.iter_func(cont)

        # Try to figure it out
        if isinstance(cont, dict):
            return cont.items()
        # Includes lists, tuples, sets etc
        return enumerate(cont)

    def _get_container(self, obj):
        if self.use_attr:
            return getattr(obj, self.use_attr)
        return obj

    def is_visitable(self, obj):
        if len(self.visit_funcs) == 0:
            cont = self._get_container(obj)
            return isinstance(cont, Iterable)
        return all(
            func(obj)
            for func in self.visit_funcs
        )




class ParserPicker:
    "Chooses parser depending of conf type"
    def __init__(self, type_parsers, subparsers=None, key_as=None):
        self.type_parsers = type_parsers
        self.subparsers = subparsers
        self.key_as = key_as

    def __call__(self, conf, **kwargs):
        "Parse item"
        parser = self.type_parsers[type(conf)]
        return parser(conf, **kwargs)



import pytest

from rocketry.parse import parse_condition
from rocketry.conditions import (
    AlwaysTrue, AlwaysFalse,
    Not,
)

@pytest.mark.parametrize(
    "cond_str,expected",
    [
        pytest.param(
            "always true",
            AlwaysTrue(),
            id="AlwaysTrue"),

        pytest.param(
            "always false",
            AlwaysFalse(),
            id="AlwaysFalse"),

        pytest.param(
            "always true & always true",
            AlwaysTrue() & AlwaysTrue(),
            id="All"),
        pytest.param(
            "always true | always true",
            AlwaysTrue() | AlwaysTrue(),
            id="Any"),
        pytest.param(
            "~always true",
            Not(AlwaysTrue()),
            id="Not"),

        pytest.param(
            "~~always true",
            ~~AlwaysTrue(),
            id="Nested Not"),

        pytest.param(
            "~(always true)",
            ~AlwaysTrue(),
            id="Closured Not"),

        pytest.param(
            "~(always true & always true)",
            ~(AlwaysTrue() & AlwaysTrue()),
            id="All in Not"),

        pytest.param(
            "~(always true | always true)",
            ~(AlwaysTrue() | AlwaysTrue()),
            id="Any in Not"),

        pytest.param(
            "~always true | ~always true & ~always true",
            ~AlwaysTrue() | ~AlwaysTrue() & ~AlwaysTrue(),
            id="Multiple Not"),

        pytest.param(
            "(always true & always true) & (always true & always true)",
            AlwaysTrue() & AlwaysTrue() & AlwaysTrue() & AlwaysTrue(),
            id="Nested AND"),
        pytest.param(
            "(always true | always true) | (always true | always true)",
            AlwaysTrue() | AlwaysTrue() | AlwaysTrue() | AlwaysTrue(),
            id="Nested OR"),
        pytest.param(
            "always true & (always true | always true & (always true | always true))",
            AlwaysTrue() & (AlwaysTrue() | AlwaysTrue() & (AlwaysTrue() | AlwaysTrue())),
            id="Deeply nested"),
        pytest.param(
            "~always true & ~~(always true | ~always true & ~(~~always true | ~always true))",
            ~AlwaysTrue() & ~~(AlwaysTrue() | ~AlwaysTrue() & ~(~~AlwaysTrue() | ~AlwaysTrue())),
            id="Deeply nested Not"),

        pytest.param(
            "",
            AlwaysFalse(),
            id="Empty", marks=pytest.mark.xfail),
    ],
)
def test_string(cond_str, expected):
    cond = parse_condition(cond_str)
    assert cond == expected

def test_bool():
    cond = parse_condition(True)
    assert cond == AlwaysTrue()

    cond = parse_condition(False)
    assert cond == AlwaysFalse()

@pytest.mark.parametrize(
    "cond_str",
    [
        pytest.param(
            "daily between 20:00 and 23:59",
            id="daily between"),
    ],
)
def test_string_matching(cond_str):
    cond = parse_condition(cond_str)



import asyncio
import multiprocessing
import time
from rocketry.args.builtin import TerminationFlag
from rocketry.conditions.scheduler import SchedulerCycles

from rocketry.tasks import FuncTask
from rocketry.time import TimeDelta
from rocketry.conds import true
from rocketry.conditions import SchedulerStarted, TaskStarted, AlwaysTrue

def run_succeeding():
    pass

def run_succeeding_slow():
    time.sleep(20)

def run_creating_child():

    proc = multiprocessing.Process(target=run_succeeding, daemon=True)
    proc.start()

def test_creating_child(session):

    FuncTask(run_creating_child, name="task_1", start_cond=AlwaysTrue(), session=session)

    session.config.tasks_as_daemon = False
    session.config.shut_cond = (TaskStarted(task="task_1") >= 1) | ~SchedulerStarted(period=TimeDelta("1 second"))

    session.start()

    logger = session["task_1"].logger
    assert 1 == logger.filter_by(action="run").count()
    assert 1 == logger.filter_by(action="success").count()
    assert 0 == logger.filter_by(action="fail").count()

def test_limited_processes(session):

    def run_thread(flag=TerminationFlag()):
        while not flag.is_set():
            ...

    async def run_async():
        while True:
            await asyncio.sleep(0)

    def do_post_check():
        sched = session.scheduler
        assert sched.n_alive == 5 # 2 processes, 1 thread, 1 async and this
        assert not sched.has_free_processors()

        assert task_threaded.is_alive()
        assert task_threaded.is_running
        assert task_async.is_alive()
        assert task_async.is_running

        assert task1.is_alive()
        assert task2.is_alive()
        assert not task3.is_alive()

        assert task1.is_running
        assert task2.is_running
        assert not task3.is_running

    task_threaded = FuncTask(run_thread, name="threaded", priority=4, start_cond=true, execution="thread", permanent=True, session=session)
    task_async = FuncTask(run_async, name="async", priority=4, start_cond=true, execution="async", permanent=True, session=session)
    post_check = FuncTask(do_post_check, name="post_check", on_shutdown=True, execution="main", session=session)

    task1 = FuncTask(run_succeeding_slow, name="task_1", priority=3, start_cond=true, execution="process", session=session)
    task2 = FuncTask(run_succeeding_slow, name="task_2", priority=2, start_cond=true, execution="process", session=session)
    task3 = FuncTask(run_succeeding_slow, name="task_3", priority=1, start_cond=true, execution="process", session=session)

    session.config.max_process_count = 2
    session.config.instant_shutdown = True
    session.config.shut_cond = SchedulerCycles() >= 3

    session.start()

    outcome = post_check.logger.filter_by().all()[-1]
    assert outcome.action == "success", outcome.exc_text




from rocketry.core import Task

class CodeTask(Task):
    """Task to run a piece of Python code

    This task may be useful for APIs where a trusted
    user can create tasks or if you have your tasks stored,
    for example, in a database.

    Parameters
    ----------
    code : str
        Piece of Python code to execute. Variable ``return_value``
        is used as the return value of this task if set. Parameters
        are passed to the code as locals.
    **kwargs : dict
        See :class:`rocketry.core.Task`

    Warnings
    --------
        Note that it is potentially dangerous if you let the user
        to create CodeTasks to your system and you don't trust the
        user. The task can do anything and is practically impossible
        to restrict running non-safe code.

    Examples
    --------
    Simple example:

    .. code-block:: python

        CodeTask('''
        for i in range(10):
            ...
        ''', start_cond='daily')

    Parametrized example with return:

    .. code-block:: python

        CodeTask('''
        baz = foo + bar
        ...
        return_value = baz
        ''', parameters={'foo': 'a value', 'bar': 'a value'})
    """
    output_variable: str = 'return_value'
    code: str

    def execute(self, **params):
        loc = params
        glob = {}
        exec(self.code, glob, loc)
        return loc.get(self.output_variable, None)

    def get_default_name(self, **kwargs):
        raise ValueError("CodeTask must have name defined")



from rocketry.args import Task
from rocketry.tasks import FuncTask

def test_run(tmpdir, session):
    task1 = FuncTask(
        lambda : None,
        name="example 1",
        execution="main",
        session=session,
    )

    task2 = FuncTask(
        lambda : None,
        name="example 2",
        execution="main",
        session=session,
    )
    session.run("example 2")
    assert task1.status is None
    assert task2.status == "success"
    assert not task1.disabled
    assert not task2.disabled

def test_run_obey_cond(session):
    task1 = FuncTask(
        lambda : None,
        name="example 1",
        execution="main",
        session=session,
    )

    task2 = FuncTask(
        lambda : None,
        name="example 2",
        execution="main",
        session=session,
    )
    session.run("example 2", obey_cond=True)
    assert task1.status is None
    assert task2.status is None # start cond is false
    assert not task1.disabled
    assert not task2.disabled

def test_run_execution(session):
    def run_task(task=Task()):
        # The execution has been temporarily set as "main"
        assert task.execution == "main"
        assert task.is_alive_as_main()
        assert not task.is_alive_as_process()
    task1 = FuncTask(
        lambda : None,
        name="example 1",
        execution="main",
        session=session,
    )

    task2 = FuncTask(
        run_task,
        name="example 2",
        execution="process",
        session=session,
    )
    # Should crash in pickling if ran with execution process
    session.run("example 2", execution="main")
    assert task1.status is None
    assert task2.status == "success"
